"""
Tools used to parallelize sections of python code.
"""
from __future__ import annotations

import os

from monty.collections import dict2namedtuple


_MAX_NPROCS = os.cpu_count()


def get_max_nprocs() -> int:
    """
    Return the maximum number of procs that can be used by AbiPy.
    """
    return _MAX_NPROCS


def set_max_nprocs(max_nprocs: int | None) -> int:
    """
    Set the maximum number of procs that can be used.
    If max_nprocs is None, os.cpu_count() is used.
    """
    global _MAX_NPROCS
    if max_nprocs is None:
        _MAX_NPROCS = os.cpu_count()
    else:
        _MAX_NPROCS = max_nprocs

    return _MAX_NPROCS


def pool_nprocs_pmode(nprocs: int | None, pmode: str):
    """
    Helper function that allows one to switch from ThreadPool to multiprocessing Pool.
    Returns named tuple with (nprocs, pool_class, using_msg)

    Args:
        nprocs: Number of procs to use. If None, use cpu_count.
        pmode: "threads": for ThreadPool.
               "processes" for multiprocessing Pool.
               "seq" for sequential execution (debugging)
    """
    from multiprocessing.pool import ThreadPool
    from multiprocessing import Pool

    max_nprocs = get_max_nprocs()

    if pmode == "seq":
        nprocs, pool_cls = 1, ThreadPool

    elif pmode == "threads":
        pool_cls = ThreadPool
        if nprocs is not None:
            pool_cls = ThreadPool if nprocs > 0 else Pool
            nprocs = min(abs(nprocs), max_nprocs)

    elif pmode == "processes":
        pool_cls = Pool
        if nprocs is not None:
            pool_cls = Pool if nprocs > 0 else ThreadPool
            nprocs = min(abs(nprocs), max_nprocs)

    else:
        raise ValueError(f"Invalid value of {pmode=}, it should be in ['seq', 'threads', 'processes']")

    return dict2namedtuple(nprocs=nprocs or max_nprocs,
                           pool_cls=pool_cls,
                           using_msg=f"using {nprocs=} with {pmode=} and Pool class: {pool_cls.__name__} ...",
                           )



#!/usr/bin/env python
r"""
Multiple Structural relaxations
===============================

This example shows how to analyze the results of multiple
structure relaxations with the HIST robot.
"""
from abipy import abilab
import abipy.data as abidata

files = [
    abidata.ref_file("sic_relax_HIST.nc"),
    abidata.ref_file("sic_relax_HIST.nc"),
]

# Build HIST robot from paths, use user-selected labels.
robot = abilab.HistRobot.from_files(files, labels=["sic_relaxation", "same_hist"])

# Get dataframe with most important results.
print(robot.get_dataframe())

robot.gridplot()

# Use combiplot to compare the results
# (meaningless in this case as we are using the same HIST file).
robot.combiplot()



# coding: utf-8
"""
This module defines the Robot BaseClass. Robots operates on multiple files and provide helper
functions to plot the data e.g. convergence studies and to build pandas dataframes from the output files.
"""
from __future__ import annotations

import sys
import os
import inspect
import itertools
import json
import numpy as np
import pandas as pd

from collections import OrderedDict, deque
from typing import Callable, Union, Any
from functools import wraps
from monty.string import is_string, list_strings
from monty.termcolor import cprint
from monty.json import MontyEncoder
from abipy.tools.serialization import pmg_serialize
from abipy.tools.iotools import make_executable
from abipy.core.structure import Structure
from abipy.core.mixins import NotebookWriter
from abipy.tools.numtools import sort_and_groupby
from abipy.tools import duck
from abipy.tools.typing import Figure
from abipy.tools.plotting import (plot_xy_with_hue, add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt,
    rotate_ticklabels, set_visible, ConvergenceAnalyzer)


class Robot(NotebookWriter):
    """
    This is the base class from which all Robot subclasses should derive.
    A Robot supports the `with` context manager:

    Usage example:

    .. code-block:: python

        with Robot([("label1", "file1"), (label2, "file2")]) as robot:
            # Do something with robot. files are automatically closed when we exit.
            for label, abifile in self.items():
                print(label)
    """
    # filepaths are relative to `start`. None for asbolute paths. This flag is set in trim_paths
    start = None

    # Used in iter_lineopt to generate matplotlib linestyles.
    _LINE_COLORS = ["b", "r", "g", "m", "y", "k", "c"]
    _LINE_STYLES = ["-", ":", "--", "-.",]
    _LINE_WIDTHS = [2, ]

    # matplotlib option to fill convergence window.
    HATCH = "/"

    def __init__(self, *args):
        """
        Args:
            args is a list of tuples (label, filepath)
        """
        self._abifiles, self._do_close = OrderedDict(), OrderedDict()
        self._exceptions = deque(maxlen=100)

        for label, abifile in args:
            self.add_file(label, abifile)

    @classmethod
    def get_supported_extensions(self) -> list[str]:
        """List of strings with extensions supported by Robot subclasses."""
        # This is needed to have all subclasses.
        from abipy.abilab import Robot
        return sorted([cls.EXT for cls in Robot.__subclasses__()])

    @classmethod
    def class_for_ext(cls, ext: str):
        """Return the Robot subclass associated to the given extension."""
        for subcls in cls.__subclasses__():
            if subcls.EXT in (ext, ext.upper()):
                return subcls

        # anaddb.nc does not follow the extension rule...
        if ext.lower() == "anaddb":
            from abipy.dfpt.anaddbnc import AnaddbNcRobot as subcls
            return subcls

        raise ValueError("Cannot find Robot subclass associated to extension %s\n" % ext +
                         "The list of supported extensions (case insensitive) is:\n%s" %
                         str(cls.get_supported_extensions()))

    @classmethod
    def from_dir(cls, top: str, walk: bool = True, abspath: bool = False) -> Robot:
        """
        Build a robot by scanning all files located within directory `top`.
        This method should be invoked with a concrete robot class, for example:

            robot = GsrRobot.from_dir(".")

        Args:
            top: Root directory
            walk: if True, directories inside `top` are included as well.
            abspath: True if paths in index should be absolute. Default: Relative to `top`.
        """
        new = cls(*cls._open_files_in_dir(top, walk))
        if not abspath: new.trim_paths(start=top)
        return new

    @classmethod
    def from_dirs(cls, dirpaths: list[str], walk: bool = True, abspath: bool = False) -> Robot:
        """
        Similar to `from_dir` but accepts a list of directories instead of a single directory.

        Args:
            walk: if True, directories inside `top` are included as well.
            abspath: True if paths in index should be absolute. Default: Relative to `top`.
        """
        items = []
        for top in list_strings(dirpaths):
            items.extend(cls._open_files_in_dir(top, walk))
        new = cls(*items)
        if not abspath: new.trim_paths(start=os.getcwd())
        return new

    @classmethod
    def from_dir_glob(cls, pattern: str, walk: bool = True, abspath: bool = False) -> Robot:
        """
        This class method builds a robot by scanning all files located within the directories
        matching `pattern` as implemented by glob.glob
        This method should be invoked with a concrete robot class, for example:

            robot = GsrRobot.from_dir_glob("flow_dir/w*/outdata/")

        Args:
            pattern: Pattern string.
            walk: if True, directories inside `top` are included as well.
            abspath: True if paths in index should be absolute. Default: Relative to getcwd().
        """
        import glob
        items = []
        for top in filter(os.path.isdir, glob.iglob(pattern)):
            items += cls._open_files_in_dir(top, walk=walk)
        new = cls(*items)
        if not abspath: new.trim_paths(start=os.getcwd())
        return new

    @classmethod
    def _open_files_in_dir(cls, top: str, walk: bool) -> list:
        """
        Open files in directory tree starting from `top`. Return list of Abinit files.
        """
        if not os.path.isdir(top):
            raise ValueError("%s: no such directory" % str(top))
        from abipy.abilab import abiopen
        items = []
        if walk:
            for dirpath, dirnames, filenames in os.walk(top):
                filenames = sorted([f for f in filenames if cls.class_handles_filename(f)])
                for f in filenames:
                    abifile = abiopen(os.path.join(dirpath, f))
                    if abifile is not None: items.append((abifile.filepath, abifile))
        else:
            filenames = [f for f in os.listdir(top) if cls.class_handles_filename(f)]
            for f in filenames:
                abifile = abiopen(os.path.join(top, f))
                if abifile is not None: items.append((abifile.filepath, abifile))

        return items

    @classmethod
    def class_handles_filename(cls, filename: str) -> bool:
        """
        True if robot class handles filename.
        """
        # Special treatment of AnaddbNcRobot
        if cls.EXT == "anaddb" and os.path.basename(filename).lower() == "anaddb.nc":
            return True

        return (filename.endswith("_" + cls.EXT + ".nc") or
                filename.endswith("." + cls.EXT))  # This for .abo

    @classmethod
    def from_files(cls, filenames, labels=None, abspath=False) -> Robot:
        """
        Build a Robot from a list of `filenames`.
        If labels is None, labels are automatically generated from absolute paths.

        Args:
            abspath: True if paths in index should be absolute. Default: Relative to `top`.
        """
        filenames = list_strings(filenames)
        from abipy.abilab import abiopen
        filenames = [f for f in filenames if cls.class_handles_filename(f)]
        items = []
        for i, f in enumerate(filenames):
            try:
                abifile = abiopen(f)
            except Exception as exc:
                cprint("Exception while opening file: `%s`" % str(f), "red")
                cprint(exc, "red")
                abifile = None

            if abifile is not None:
                label = abifile.filepath if labels is None else labels[i]
                items.append((label, abifile))

        new = cls(*items)
        if labels is None and not abspath: new.trim_paths(start=None)
        return new

    @classmethod
    def from_json_file(cls, filepath: str):
        """
        Build Robot from a file in json format.
        """
        from abipy.tools.serialization import mjson_load
        new = mjson_load(filepath)
        return new

    @classmethod
    def from_top_and_json_basename(cls, top: str, json_basename: str) -> Robot:
        """
        Build a robot by scanning all json files located within directory `top`.
        and matching json_basename.

        Example:

            gwr_robot = Robot.from_top_and_json_basename(".", "gwr_robot.json")
        """
        json_paths = []
        for root, dirs, files in os.walk(top):
            for name in files:
                if name == json_basename:
                    json_paths.append(os.path.join(root, name))

        return cls.from_json_files(json_paths)

    @classmethod
    def from_json_files(cls, json_paths: list[str]):
        """
        Build Robot from a list of json files.
        Each json file should have a list of filepaths and @module and @class as required by msonable.
        """
        # Merge filepaths and chech that module and class are equal across files.
        robot_filepaths = []
        _module, _class = None, None
        for path in json_paths:
            with open(path, "rt") as fh:
                d = json.load(fh)

            if "filepaths" not in d:
                raise ValueError(f"{path=} does not provide `filepaths` entry")

            # Merge filepaths
            robot_filepaths.extend(d["filepaths"])

            if _module is None:
                _module, _class = d["@module"], d["@class"]
            else:
                if _module != d["@module"]:
                    raise ValueError(f'{_module=} != {d["@module"]=}')
                if _class != d["@class"]:
                    raise ValueError(f'{_class=} != {d["@class"]=}')

        # Build Robot instance from string in json format.
        #print("Merged filepaths:", robot_filepaths)
        from abipy.tools.serialization import mjson_loads
        data = {"filepaths": robot_filepaths, "@class": _class, "@module": _module}
        return mjson_loads(json.dumps(data))

    @classmethod
    def from_work(cls, work, outdirs="all", nids=None, ext=None, task_class=None) -> Robot:
        """
        Build a robot from a |Work| object.
        """
        nids = [work.node_id] + [task.node_id for task in work]
        return cls.from_flow(work.flow, nids=nids)

    @classmethod
    def from_flow(cls, flow, outdirs="all", nids=None, ext=None, task_class=None) -> Robot:
        """
        Build a robot from a |Flow| object.

        Args:
            flow: |Flow| object
            outdirs: String used to select/ignore the files in the output directory of flow, works and tasks
                outdirs="work" selects only the outdir of the Works,
                outdirs="flow+task" selects the outdir of the Flow and the outdirs of the tasks
                outdirs="-work" excludes the outdir of the Works.
                Cannot use ``+`` and ``-`` flags in the same string.
                Default: `all` that is equivalent to "flow+work+task"
            nids: List of node identifiers used to select particular nodes. Not used if None
            ext: File extension associated to the robot. Mainly used if method is invoked with the BaseClass
            task_class: Task class or string with the class name used to select the tasks in the flow.
                None implies no filtering.

        Usage example:

        .. code-block:: python

            with abilab.GsrRobot.from_flow(flow) as robot:
                print(robot)

        Returns:
            ``Robot`` subclass.
        """
        robot = cls() if ext is None else cls.class_for_ext(ext)()
        all_opts = ("flow", "work", "task")

        if outdirs == "all":
            tokens = all_opts
        elif "+" in outdirs:
            assert "-" not in outdirs
            tokens = outdirs.split("+")
        elif "-" in outdirs:
            assert "+" not in outdirs
            tokens = [s for s in all if s not in outdirs.split("-")]
        else:
            tokens = list_strings(outdirs)

        if not all(t in all_opts for t in tokens):
            raise ValueError("Wrong outdirs string %s" % outdirs)

        if "flow" in tokens:
            robot.add_extfile_of_node(flow, nids=nids, task_class=task_class)

        if "work" in tokens:
            for work in flow:
                robot.add_extfile_of_node(work, nids=nids, task_class=task_class)

        if "task" in tokens:
            for task in flow.iflat_tasks():
                robot.add_extfile_of_node(task, nids=nids, task_class=task_class)

        return robot


    def __len__(self):
        return len(self._abifiles)

    #def __iter__(self):
    #    return iter(self._abifiles)

    def __getitem__(self, key):
        # self[key]
        return self._abifiles.__getitem__(key)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Activated at the end of the with statement."""
        self.close()

    def keys(self):
        return self._abifiles.keys()

    def items(self):
        return self._abifiles.items()

    def add_extfile_of_node(self, node, nids=None, task_class=None) -> None:
        """
        Add the file produced by this node to the robot.

        Args:
            node: |Flow| or |Work| or |Task| object.
            nids: List of node identifiers used to select particular nodes. Not used if None
            task_class: Task class or string with class name used to select the tasks in the flow.
                None implies no filtering.
        """
        if nids and node.node_id not in nids: return
        filepath = node.outdir.has_abiext(self.EXT)
        if not filepath:
            # Look in run.abi directory.
            filepath = node.wdir.has_abiext(self.EXT)

        # This to ignore DDB.nc files (only text DDB are supported)
        if filepath and filepath.endswith("_DDB.nc"):
            return

        if filepath:
            try:
                label = os.path.relpath(filepath)
            except OSError:
                # current working directory may not be defined!
                label = filepath

            # Filter by task_class (class or string with class name)
            if task_class is not None and not node.isinstance(task_class):
                return None

            self.add_file(label, filepath)

    def scan_dir(self, top: str, walk: bool = True) -> int:
        """
        Scan directory tree starting from ``top``.
        Add files to the robot instance.
        Return: Number of files found.

        Args:
            top: Root directory
            walk: if True, directories inside ``top`` are included as well.
        """
        count = 0
        for filepath, abifile in self.__class__._open_files_in_dir(top, walk):
            count += 1
            self.add_file(filepath, abifile)

        return count

    def add_file(self, label, abifile, filter_abifile=None) -> None:
        """
        Add a file to the robot with the given label.

        Args:
            label: String used to identify the file (must be unique, ax exceptions is
                raised if label is already present.
            abifile: Specify the file to be added. Accepts strings (filepath) or abipy file-like objects.
            filter_abifile: Function that receives an ``abifile`` object and returns
                True if the file should be added to the plotter.
        """
        if is_string(abifile):
            from abipy.abilab import abiopen
            abifile = abiopen(abifile)
            if filter_abifile is not None and not filter_abifile(abifile):
                abifile.close()
                return

            # Open file here --> have to close it.
            self._do_close[abifile.filepath] = True

        if label in self._abifiles:
            raise ValueError("label %s is already present!" % label)

        self._abifiles[label] = abifile

    @pmg_serialize
    def as_dict(self) -> dict:
        """Return dict with filepaths that can be used to reconstruct the Robot."""
        return dict(filepaths=[abifile.filepath for abifile in self.abifiles])

    @classmethod
    def from_dict(cls, d: dict):
        """Recontruct object from dictionary with filepaths."""
        return cls.from_files(d["filepaths"])

    def to_json(self) -> str:
        """
        Returns a JSON string representation of the object.
        """
        return json.dumps(self.as_dict(), cls=MontyEncoder)

    def get_pyscript(self, filepath: str) -> RobotPythonScript:
        """Return RobotPythonScript to br used as context manager."""
        return RobotPythonScript(self, filepath)

    #def pop_filepath(self, filepath: str) -> None:
    #    """
    #    Remove the file with the given `filepath`.
    #    """
    #    if label, abifile in self._abifiles.items():
    #        if abifile.filepath != filepath: continue
    #        self._abifiles.pop(label)

    def iter_lineopt(self):
        """Generates matplotlib linestyles."""
        for o in itertools.product( self._LINE_WIDTHS,  self._LINE_STYLES, self._LINE_COLORS):
            yield {"linewidth": o[0], "linestyle": o[1], "color": o[2]}

    @staticmethod
    def ordered_intersection(list_1, list_2) -> list:
        """Return ordered intersection of two lists. Items must be hashable."""
        set_2 = frozenset(list_2)
        return [x for x in list_1 if x in set_2]

    #def _get_ointersection_i(self, iattrname):
    #    if len(self.abifiles) == 0: return []
    #    values = list(range(getattr(self.abifiles[0], iattrname)))
    #    if len(self.abifiles) == 1: return values
    #    for abifile in self.abifiles[1:]:
    #        values = self.ordered_intersection(values, range(getattr(abifile, iattrname)))
    #    return values

    @staticmethod
    def _to_relpaths(paths) -> list[str]:
        """Convert a list of absolute paths to relative paths."""
        root = os.getcwd()
        return [os.path.relpath(p, root) for p in paths]

    def remove(self) -> None:
        """Close the file handle, remove the file from disk for each file in the robot."""
        for abifile in self.abifiles:
            abifile.remove()

    def pop_label(self, label: str) -> None:
        """
        Remove file with the given ``label`` and close it.
        """
        if label in self._abifiles:
            abifile = self._abifiles.pop(label)
            if self._do_close.pop(abifile.filepath, False):
                try:
                    abifile.close()
                except Exception as exc:
                    print("Exception while closing: ", abifile.filepath)
                    print(exc)

    def change_labels(self, new_labels: list[str], dryrun: bool = False) -> dict:
        """
        Change labels of the files.

        Args:
            new_labels: List of strings (same length as self.abifiles)
            dryrun: True to activate dryrun mode.

        Return:
            mapping new_label --> old_label.
        """
        if len(new_labels) != len(self):
            raise ValueError("Robot has %d files while len(new_labels) = %d" % (len(new_labels), len(self)))

        old_labels = list(self._abifiles.keys())
        if not dryrun:
            old_abifiles, self._abifiles = self._abifiles, OrderedDict()
        new2old = OrderedDict()
        for old, new in zip(old_labels, new_labels):
            new2old[new] = old
            if not dryrun:
                self._abifiles[new] = old_abifiles[old]
            else:
                print("old [%s] --> new [%s]" % (old, new))

        return new2old

    def remap_labels(self, function: Callable, dryrun: bool = False) -> dict:
        """
        Change labels of the files by executing ``function``

        Args:
            function: Callable object e.g. lambda function. The output of function(abifile) is used as
                new label. Note that the function shall not return duplicated labels when applied to self.abifiles.
            dryrun: True to activate dryrun mode.

        Return:
            mapping new_label --> old_label.
        """
        new_labels = [function(afile) for afile in self.abifiles]
        # Labels must be unique and hashable.
        if len(set(new_labels)) != len(new_labels):
            raise ValueError("Duplicated labels are not allowed. Change input function.\nnew_labels %s" % str(new_labels))

        return self.change_labels(new_labels, dryrun=dryrun)

    def trim_paths(self, start=None) -> str:
        """
        Replace absolute filepaths in the robot with relative paths wrt to ``start`` directory.
        If start is None, os.getcwd() is used. Set ``self.start`` attribute, return ``self.start``.
        """
        self.start = os.getcwd() if start is None else start
        old_paths = list(self._abifiles.keys())
        old_new_paths = [(p, os.path.relpath(os.path.abspath(p), start=self.start)) for p in old_paths]

        old_abifiles = self._abifiles
        self._abifiles = OrderedDict()
        for old, new in old_new_paths:
            self._abifiles[new] = old_abifiles[old]

        return self.start

    @property
    def exceptions(self) -> list:
        """List of exceptions."""
        return self._exceptions

    @property
    def labels(self) -> list[str]:
        """
        List of strings used to create labels in matplotlib figures when plotting results
        taked from multiple files. By default, labels is initialized with the path of the files in the robot.
        Use change_labels to change the list.
        """
        return list(self._abifiles.keys())

    def get_label_files_str(self):
        """Return string with [label, filepath]."""
        from tabulate import tabulate
        return tabulate([(label, abifile.relpath) for label, abifile in self.items()],
                        headers=["Label", "Relpath"]) + "\n"

    def show_files(self, stream=sys.stdout) -> None:
        """Show label --> file path"""
        stream.write(self.get_label_files_str())

    def __repr__(self):
        """Invoked by repr."""
        return self.get_label_files_str()

    def __str__(self):
        """Invoked by str."""
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = ["%s with %d files in memory:\n" % (self.__class__.__name__, len(self.abifiles))]
        app = lines.append
        for i, f in enumerate(self.abifiles):
            app(f.to_string(verbose=verbose))
            app("\n")

        return "\n".join(lines)

    def _repr_html_(self) -> str:
        """Integration with jupyter_ notebooks."""
        return '<ol start="0">\n{}\n</ol>'.format("\n".join("<li>%s</li>" % label for label, abifile in self.items()))

    def getattr_alleq(self, aname : str):
        """
        Return the value of attribute aname.
        Raises ValueError if value is not the same across all the files in the robot.
        """
        val1 = getattr(self.abifiles[0], aname)

        for abifile in self.abifiles[1:]:
            val2 = getattr(abifile, aname)
            if isinstance(val1, (str, int, float)):
                eq = val1 == val2
            elif isinstance(val1, np.ndarray):
                eq = np.allclose(val1, val2)
            if not eq:
                raise ValueError(f"Different values of {aname=}, {val1=}, {val2=}")

        return val1

    @property
    def abifiles(self) -> list:
        """List of netcdf files."""
        return list(self._abifiles.values())

    def has_different_structures(self, rtol=1e-05, atol=1e-08) -> str:
        """
        Check if structures are equivalent, return string with info about differences (if any).
        """
        if len(self) <= 1: return ""
        formulas = set([af.structure.composition.formula for af in self.abifiles])
        if len(formulas) != 1:
            return "Found structures with different full formulas: %s" % str(formulas)

        lines = []
        s0 = self.abifiles[0].structure
        for abifile in self.abifiles[1:]:
            s1 = abifile.structure
            if not np.allclose(s0.lattice.matrix, s1.lattice.matrix, rtol=rtol, atol=atol):
                lines.append("Structures have different lattice:")
            if not np.allclose(s0.frac_coords, s1.frac_coords, rtol=rtol, atol=atol):
                lines.append("Structures have different atomic positions:")

        return "\n".join(lines)

    def _get_ref_abifile_from_basename(self, ref_basename: str | None):
        """
        Find reference abifile. If None, the first file in the robot is used.
        """
        ref_file = self.abifiles[0]
        if ref_basename is None:
            return ref_file

        for i, abifile in enumerate(self.abifiles):
            if abifile.basename == ref_basename:
                return abifile

        raise ValueError(f"Cannot find {ref_basename=}")

    @staticmethod
    def _compare_attr_name(aname: str, ref_abifile, other_abifile) -> None:
        """
        Compare the value of attribute `aname` in two files.
        """
        # Get attributes in abifile first, then in abifile.r, else raise.
        if hasattr(ref_abifile, aname):
            val1, val2 = getattr(ref_abifile, aname), getattr(other_abifile, aname)

        elif hasattr(ref_abifile , "r") and hasattr(ref_abifile.r, aname):
            val1, val2 = getattr(ref_abifile.r, aname), getattr(other_abifile.r, aname)

        else:
            raise AttributeError(f"Cannot find attribute `{aname =}`")

        # Now compare val1 and val2 taking into account the type.
        if isinstance(val1, (str, int, float, Structure)):
            eq = val1 == val2
        elif isinstance(val1, np.ndarray):
            eq = np.allclose(val1, val2)
        else:
            raise TypeError(f"Don't know how to handle comparison for type: {type(val1)}")

        if not eq:
            raise ValueError(f"Different values of {aname=}, {val1=}, {val2=}")

    def is_sortable(self, aname: str, raise_exc: bool = False) -> bool:
        """
        Return True if ``aname`` is an attribute of the netcdf file
        If raise_exc is True, AttributeError with an explicit message is raised.
        """
        try:
            obj = None
            try:
                # abifile.foo.bar ?
                obj = duck.getattrd(self.abifiles[0], aname)
            except AttributeError:
                # abifile.params[aname] ?
                if hasattr(self.abifiles[0], "params") and aname in self.abifiles[0].params:
                    obj = self.abifiles[0].params[aname]

            # Let's try to convert obj to scalar.
            float(obj)
            return True

        except Exception:
            if not raise_exc: return False
            attrs = []
            for key, obj in inspect.getmembers(self.abifiles[0]):
                # Ignores anything starting with underscore
                if key.startswith('_') or callable(obj) or hasattr(obj, "__len__"): continue
                attrs.append(key)

            # Add all keys in params dict.
            if hasattr(self.abifiles[0], "params") and hasattr(self.abifiles[0].params, "keys"):
                attrs.extend(self.abifiles[0].params.keys())

            raise AttributeError("""\
`%s` object has no attribute `%s`. Choose among:

    %s

Note that this list is automatically generated.
Not all entries are sortable (Please select number-like quantities)""" % (self.__class__.__name__, aname, str(attrs)))

    def _sortby_labelfile_list(self, labelfile_list, func_or_string, reverse=False, unpack=False):
        """
        Return: list of (label, abifile, xs) tuples where xs is obtained via ``func_or_string``.
            or labels, abifiles, xs if ``unpack``
        """
        if not func_or_string:
            # Catch None or empty
            items = [(label, abifile, label) for (label, abifile) in labelfile_list]
            if not unpack:
                return items
            else:
                return [t[0] for t in items], [t[1] for t in items], [t[2] for t in items]

        elif callable(func_or_string):
            items = [(label, abifile, func_or_string(abifile)) for (label, abifile) in labelfile_list]

        else:
            # Assume string and attribute with the same name.
            # try in abifile.params if not hasattrd(abifile, func_or_string)
            self.is_sortable(func_or_string, raise_exc=True)
            if duck.hasattrd(self.abifiles[0], func_or_string):
                items = [(label, abifile, duck.getattrd(abifile, func_or_string)) for (label, abifile) in labelfile_list]

            #elif hasattr(self.abifiles[0], "reader") and duck.hasattrd(self.abifiles[0].reader, func_or_string):
            #    items = [(label, abifile, duck.getattrd(abifile.reader, func_or_string)) for (label, abifile) in labelfile_list]

            #elif hasattr(self.abifiles[0], "r") and duck.hasattrd(self.abifiles[0].r, func_or_string):
            #elif hasattr(self.abifiles[0], "r"):
            #    items = [(label, abifile, duck.getattrd(abifile.r, func_or_string)) for (label, abifile) in labelfile_list]

            else:
                #print("Cannot find", func_or_string, " in self.abifiles[0]")
                items = [(label, abifile, abifile.params[func_or_string]) for (label, abifile) in labelfile_list]

        items = sorted(items, key=lambda t: t[2], reverse=reverse)
        if not unpack:
            return items
        else:
            return [t[0] for t in items], [t[1] for t in items], [t[2] for t in items]

    def sortby(self, func_or_string: Union[Callable, str, None],
               reverse: bool = False, unpack: bool = False) -> list[tuple]:
        """
        Sort files in the robot by ``func_or_string``.

        Args:
            func_or_string: Can be None, string or callable defining the quantity to be used for sorting.
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of func_or_string(abifile) is used.
                If None, no sorting is performed.
            reverse: If set to True, then the list elements are sorted as if each comparison were reversed.
            unpack: Return (labels, abifiles, xs) if True

        Return: list of (label, abifile, xs) tuples where xs is obtained via ``func_or_string``.
            or labels, abifiles, xs if ``unpack``
        """
        labelfile_list = list(self.items())
        return self._sortby_labelfile_list(labelfile_list, func_or_string, reverse=reverse, unpack=unpack)

    def group_and_sortby(self,
                         hue: Union[Callable, str],
                         func_or_string: Union[Callable, str, None]) -> list[HueGroup]:
        """
        Group files by ``hue`` and, inside each group` sort items by ``func_or_string``.

        Args:
            hue: Variable that defines subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                Dot notation is also supported e.g. hue="structure.formula" --> abifile.structure.formula
                If callable, the output of hue(abifile) is used.
            func_or_string: Either None, string, callable defining the quantity to be used for sorting.
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of func_or_string(abifile) is used.
                If None, no sorting is performed.

        Return: List of :class:`HueGroup` instance.
        """
        # Group by hue.
        # This is the section in which we support: callable, abifile.attr.name syntax or abifile.params["key"]
        items = list(self.items())

        if callable(hue):
            key = lambda t: hue(t[1])
        else:
            # Assume string.
            if duck.hasattrd(self.abifiles[0], hue):
                key = lambda t: duck.getattrd(t[1], hue)
            else:
                # Try in abifile.params
                if hasattr(self.abifiles[0], "params") and hue in self.abifiles[0].params:
                    key = lambda t: t[1].params[hue]
                else:
                    raise TypeError("""\
Cannot interpret hue argument of type `%s` and value `%s`.
Expecting callable or attribute name or key in abifile.params""" % (type(hue), str(hue)))

        groups = []
        for hvalue, labelfile_list in sort_and_groupby(items, key=key):
            # Use func_or_string to sort each group
            labels, abifiles, xvalues = self._sortby_labelfile_list(labelfile_list, func_or_string, unpack=True)
            groups.append(HueGroup(hvalue, xvalues, abifiles, labels))

        return groups

    def close(self) -> None:
        """
        Close all files that have been opened by the Robot.
        """
        for abifile in self.abifiles:
            if self._do_close.pop(abifile.filepath, False):
                try:
                    abifile.close()
                except Exception as exc:
                    print("Exception while closing: ", abifile.filepath)
                    print(exc)

    def _exec_funcs(self, funcs, arg) -> dict:
        """
        Execute list of callable functions. Each function receives arg as argument.
        """
        if not isinstance(funcs, (list, tuple)): funcs = [funcs]
        d = {}
        for func in funcs:
            try:
                key, value = func(arg)
                d[key] = value
            except Exception as exc:
                cprint("Exception: %s" % str(exc), "red")
                self._exceptions.append(str(exc))
        return d

    @staticmethod
    def sortby_label(sortby, param) -> str:
        """Return the label to be used when files are sorted with ``sortby``."""
        return "%s %s" % (sortby, param) if not (callable(sortby) or sortby is None) else str(param)

    def get_structure_dataframes(self, abspath=False, filter_abifile=None, **kwargs):
        """
        Wrap dataframes_from_structures function.

        Args:
            abspath: True if paths in index should be absolute. Default: Relative to getcwd().
            filter_abifile: Function that receives an ``abifile`` object and returns
                True if the file should be added to the plotter.
        """
        from abipy.core.structure import dataframes_from_structures
        if "index" not in kwargs:
            index = list(self._abifiles.keys())
            if not abspath: index = self._to_relpaths(index)
            kwargs["index"] = index

        abifiles = self.abifiles if filter_abifile is not None else list(filter(filter_abifile, self.abifiles))
        return dataframes_from_structures(struct_objects=abifiles, **kwargs)

    def get_lattice_dataframe(self, **kwargs) -> pd.DataFrame:
        """Return |pandas-DataFrame| with lattice parameters."""
        dfs = self.get_structure_dataframes(**kwargs)
        return dfs.lattice

    def get_coords_dataframe(self, **kwargs) -> pd.DataFrame:
        """Return |pandas-DataFrame| with atomic positions."""
        dfs = self.get_structure_dataframes(**kwargs)
        return dfs.coords

    def get_params_dataframe(self, abspath: bool = False) -> pd.DataFrame:
        """
        Return |pandas-DataFrame| with the most important parameters.
        that are usually subject to convergence studies.

        Args:
            abspath: True if paths in index should be absolute. Default: Relative to `top`.
        """
        rows, row_names = [], []
        for label, abifile in self.items():
            if not hasattr(abifile, "params"):
                import warnings
                warnings.warn("%s does not have `params` attribute" % type(abifile))
                break
            rows.append(abifile.params)
            row_names.append(label)

        row_names = row_names if abspath else self._to_relpaths(row_names)
        return pd.DataFrame(rows, index=row_names, columns=list(rows[0].keys()))

    ##############################################
    # Helper functions to plot pandas dataframes #
    ##############################################

    @staticmethod
    @wraps(plot_xy_with_hue)
    def plot_xy_with_hue(*args, **kwargs) -> Figure:
        return plot_xy_with_hue(*args, **kwargs)

    @staticmethod
    def _get_label(func_or_string) -> str:
        """
        Return label associated to ``func_or_string``.
        If callable, docstring __doc__ is used.
        """
        if func_or_string is None:
            return ""
        elif callable(func_or_string):
            if getattr(func_or_string, "__doc__", ""):
                return func_or_string.__doc__.strip()
            else:
                return func_or_string.__name__
        else:
            return str(func_or_string)

    @add_fig_kwargs
    def plot_convergence(self, item: Union[str, Callable],
                         sortby=None, hue=None, abs_conv=None,
                         ax=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of ``item`` wrt the ``sortby`` parameter.
        Values can optionally be grouped by ``hue``.

        Args:
            item: Define the quantity to plot. Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and `getattr` is invoked.
                Dot notation is also supported e.g. hue="structure.formula" --> `abifile.structure.formula`.
                If callable, the output of `item(abifile)` is used.
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string.
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of hue(abifile) is used.
            abs_conv: If not None, plot f(x) and abs(f(x) - f(x_inf)) in log scale.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: legend and label fontsize.
            kwargs: keyword arguments passed to matplotlib plot method.

        Returns: |matplotlib-Figure|

        Example:

             robot.plot_convergence("energy")
             robot.plot_convergence("energy", sortby="nkpt")
             robot.plot_convergence("pressure", sortby="nkpt", hue="tsmear")
             robot.plot_convergence("pressure", sortby="nkpt", hue="tsmear", abs_conv=1e-3)
        """
        if "marker" not in kwargs:
            kwargs["marker"] = "o"

        if abs_conv is None:
            # Plot f(x) only
            ax1, fig, plt = get_ax_fig_plt(ax=ax)
            ax2 = None
        else:
            # Plot f(x) and |f(x) - f(x_inf)| in log scale.
            (ax1, ax2), fig, plt = get_axarray_fig_plt(ax, nrows=2, ncols=1,
                                                       sharex=True, sharey=False, squeeze=True)
            ax2.grid(True)

        if hue is None:
            labels, abifiles, xs = self.sortby(sortby, unpack=True)
            yvals = self.get_yvals_item_abifiles(item, abifiles)
            ax1.plot(xs, yvals, **kwargs)

            if ax2:
                self.plot_abs_conv(ax1, ax2, xs, yvals, abs_conv, self._get_label(sortby),
                                   fontsize, self.HATCH, **kwargs)

        else:
            groups = self.group_and_sortby(hue, sortby)
            for g in groups:
                yvals = self.get_yvals_item_abifiles(item, g.abifiles)
                label = "%s: %s" % (self._get_label(hue), g.hvalue)
                ax1.plot(g.xvalues, yvals, label=label, **kwargs)

                if ax2:
                    self.plot_abs_conv(ax1, ax2, g.xvalues, yvals, abs_conv, self._get_label(sortby),
                                       fontsize, self.HATCH, **kwargs)

        ax1.grid(True)
        ax1.set_xlabel("%s" % self._get_label(sortby))
        if sortby is None: rotate_ticklabels(ax1, 15)
        ax1.set_ylabel("%s" % self._get_label(item))

        if hue is not None:
            ax1.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    @add_fig_kwargs
    def plot_convergence_items(self, items: list[Union[str, Callable]],
                               sortby=None, hue=None, abs_conv=None,
                               fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of a list of ``items`` wrt to the ``sortby`` parameter.
        Values can optionally be grouped by ``hue``.

        Args:
            items: List of attributes (or callables) to be analyzed.
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                Dot notation is also supported e.g. hue="structure.formula" --> abifile.structure.formula
                If callable, the output of hue(abifile) is used.
            abs_conv: If not None, plot f(x) and abs(f(x) - f(x_inf)) in log scale.
                Since we are plotting multiple quantities, abs_conv is a dict mapping the name of the item to
                to the convergence.
            fontsize: legend and label fontsize.
            kwargs: keyword arguments are passed to ax.plot

        Returns: |matplotlib-Figure|
        """
        # Note: in principle one could call plot_convergence inside a loop but
        # this one is faster as sorting is done only once.

        # Build grid plot.
        nrows, ncols = len(items), 1 if abs_conv is None else 2
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)

        # Sort and group files if hue.
        if hue is None:
            labels, ncfiles, xs = self.sortby(sortby, unpack=True)
        else:
            groups = self.group_and_sortby(hue, sortby)

        if "marker" not in kwargs:
            kwargs["marker"] = "o"

        for i, (ax_row, item) in enumerate(zip(ax_mat, items)):
            ax1 = ax_row[0]
            ax2 = ax_row[1] if abs_conv is not None else None

            if hue is None:
                # Extract data.
                yvals = self.get_yvals_item_abifiles(item, self.abifiles)

                self.plot_xvals_or_xstr_ax(ax1, xs, yvals, fontsize, **kwargs)

                if ax2:
                    xlabel = self._get_label(sortby) if i == len(items) - 1 else None
                    self.plot_abs_conv(ax1, ax2, xs, yvals, abs_conv[item], xlabel, fontsize, self.HATCH,
                                       **kwargs)

            else:
                for g in groups:
                    # Extract data in group
                    yvals = self.get_yvals_item_abifiles(item, g.abifiles)

                    label = "%s: %s" % (self._get_label(hue), g.hvalue)
                    ax1.plot(g.xvalues, yvals, label=label, **kwargs)

                    if ax2:
                        self.plot_abs_conv(ax1, ax2, g.xvalues, yvals, abs_conv[item], None,
                                       fontsize, self.HATCH, **kwargs)

            ax1.grid(True)
            ax1.set_ylabel(self._get_label(item))
            if ax2:
                ax2.grid(True)

            if i == len(items) - 1:
                ax1.set_xlabel("%s" % self._get_label(sortby))
                if sortby is None: rotate_ticklabels(ax1, 15)

            if i == 0 and hue is not None:
                ax1.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    def get_convergence_analyzer(self, xname: str, ytols_dict: dict) -> ConvergenceAnalyzer:
        """
        The main difference is that ConvergenceAnalyze supports multiple convergence tolerances
        for a given y-value.

        Args:
            xname: Name of the x-variable.
            ytols_dict: dict mapping the name of the y-variable to the tolerance(s).

        Example:
        """
        df = self.get_dataframe()
        return ConvergenceAnalyzer.from_dataframe(df, xname, ytols_dict)

    @add_fig_kwargs
    def plot_lattice_convergence(self, what_list=None, sortby=None, hue=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of the lattice parameters (a, b, c, alpha, beta, gamma).
        wrt the``sortby`` parameter. Values can optionally be grouped by ``hue``.

        Args:
            what_list: List of strings with the quantities to plot e.g. ["a", "alpha", "beta"].
                None means all.
            item: Define the quantity to plot. Accepts callable or string
                If string, it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of item(abifile) is used.
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                Dot notation is also supported e.g. hue="structure.formula" --> abifile.structure.formula
                If callable, the output of hue(abifile) is used.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: legend and label fontsize.

        Returns: |matplotlib-Figure|

        Example:

             robot.plot_lattice_convergence()

             robot.plot_lattice_convergence(sortby="nkpt")

             robot.plot_lattice_convergence(sortby="nkpt", hue="tsmear")
        """
        if not self.abifiles: return None

        # The majority of AbiPy files have a structure object
        # whereas Hist.nc defines final_structure. Use geattr and key to extract structure object.
        key = "structure"
        if not hasattr(self.abifiles[0], "structure"):
            if hasattr(self.abifiles[0], "final_structure"):
                key = "final_structure"
            else:
                raise TypeError("Don't know how to extract structure from %s" % type(self.abifiles[0]))

        # Define callbacks. docstrings will be used as ylabels.
        def a(afile):
            "a (Ang)"
            return getattr(afile, key).lattice.a
        def b(afile):
            "b (Ang)"
            return getattr(afile, key).lattice.b
        def c(afile):
            "c (Ang)"
            return getattr(afile, key).lattice.c
        def volume(afile):
            r"$V$"
            return getattr(afile, key).lattice.volume
        def alpha(afile):
            r"$\alpha$"
            return getattr(afile, key).lattice.alpha
        def beta(afile):
            r"$\beta$"
            return getattr(afile, key).lattice.beta
        def gamma(afile):
            r"$\gamma$"
            return getattr(afile, key).lattice.gamma

        items = [a, b, c, volume, alpha, beta, gamma]
        if what_list is not None:
            locs = locals()
            items = [locs[what] for what in list_strings(what_list)]

        # Build plot grid.
        nrows, ncols = len(items), 1
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)

        marker = kwargs.pop("marker", "o")
        for i, (ax, item) in enumerate(zip(ax_list.ravel(), items)):
            self.plot_convergence(item, sortby=sortby, hue=hue, ax=ax, fontsize=fontsize,
                                  marker=marker, show=False)
            if i != 0:
                set_visible(ax, False, "legend")
            if i != len(items) - 1:
                set_visible(ax, False, "xlabel")

        return fig

    def get_baserobot_code_cells(self, title=None) -> list:
        """
        Return list of jupyter_ cells with calls to methods provided by the base class.
        """
        # Try not pollute namespace with lots of variables.
        nbformat, nbv = self.get_nbformat_nbv()
        title = "## Code to compare multiple Structure objects" if title is None else str(title)
        return [
            nbv.new_markdown_cell(title),
            nbv.new_code_cell("robot.get_lattice_dataframe()"),
            nbv.new_code_cell("""# robot.plot_lattice_convergence(sortby="nkpt", hue="tsmear")"""),
            nbv.new_code_cell("#robot.get_coords_dataframe()"),
        ]

    ##########################################################
    # Helper functions used by Robot to extract and plot data
    ##########################################################

    @staticmethod
    def get_yvals_item_abifiles(item: Any, abifiles: list) -> np.ndarray:
        """Extract values for a list of Abinit files."""
        def _float(obj):
            if obj is None: return obj
            return float(obj)

        if callable(item):
            return np.array([_float(item(a)) for a in abifiles])
        else:
            return np.array([_float(duck.getattrd(a, item)) for a in abifiles])

    @staticmethod
    def plot_xvals_or_xstr_ax(ax, xs, yvals, fontsize, **kwargs) -> list:
        """Plot xs where xs can contain either numbers or strings."""
        if not is_string(xs[0]):
            lines = ax.plot(xs, yvals, **kwargs)
        else:
            # Must handle list of strings in a different way.
            xn = range(len(xs))
            lines = ax.plot(xn, yvals, **kwargs)
            ax.set_xticks(xn)
            ax.set_xticklabels(xs, fontsize=fontsize)

        return lines

    @staticmethod
    def plot_abs_conv(ax1, ax2, xs, yvals, abs_conv, xlabel, fontsize, hatch, **kwargs) -> None:
        """
        Plot |y - y_xmax| in log scale on ax2 and add hspan to ax1.
        """
        y_xmax = yvals[-1]
        span_style = dict(alpha=0.2, color="green", hatch=hatch)
        ax1.axhspan(y_xmax - abs_conv, y_xmax + abs_conv, label=r"$|y-y(x_{max})| \leq %s$" % abs_conv, **span_style)

        # Plot |y - y_xmax| in log scale on ax2.
        ax2.plot(xs, np.abs(yvals - y_xmax), **kwargs)
        ax2.set_yscale("log")
        ax2.set_ylabel(r"$|y-y_{x_{max}}|$", fontsize=fontsize)
        ax2.axhspan(0, abs_conv, label=r"$|y-y(x_{max})| \leq %s$" % abs_conv, **span_style)
        ax2.legend(loc="best", fontsize=fontsize, shadow=True)
        if xlabel:
            ax2.set_xlabel("%s" % xlabel)



class HueGroup:
    """
    This small object is used by ``group_and_sortby`` to store information about the group.
    """

    def __init__(self, hvalue, xvalues, abifiles, labels):
        """
        Args:
            hvalue: Hue value.
            xvalues: abifiles are sorted by ``func_or_string`` and these are the values
                associated to ``abifiles``.
            abifiles: List of file with this hue value.
            labels: List of labels associated to ``abifiles``.
        """
        self.hvalue = hvalue
        self.abifiles = abifiles
        self.labels = labels
        self.xvalues = xvalues
        assert len(abifiles) == len(labels)
        assert len(abifiles) == len(xvalues)

    def __len__(self) -> int:
        return len(self.abifiles)

    def __iter__(self):
        """Iterate over (label, abifile, xvalue)."""
        return zip(self.labels, self.abifiles, self.xvalues)



class RobotPythonScript:
    """
    Small object used to generate a python script that reconstructs the
    robot from a json file containing the list of files.
    Client code can then add additional logic to the script and write it to disk.

    This object is typically used in the `on_all_ok` method of Works
    to generate ready-to-use python scripts to post-process/visualize the results.

    Example:

        with gsr_robot.get_pyscript(work.outdir.path_in("gsr_robot.py")) as script:
            script.add_text("a = 1")
    """

    def __init__(self, robot: Robot, filepath_py: str):
        self.robot = robot
        self.filepath_py = filepath_py

        root, _ = os.path.splitext(filepath_py)
        self.filepath_json = root + ".json"
        basename = os.path.basename(filepath_py)

        self.pytext = f"""\
#!/usr/bin/env python
# This script has been automatically generated by AbiPy.
from __future__ import annotations

if False:
    import seaborn as sns
    sns.set(context="paper", style='darkgrid', palette='deep',
           font='sans-serif', font_scale=0.8, color_codes=False, rc=None)

from abipy.abio.robots import Robot
robot = Robot.from_json_file("{self.filepath_json}")
print(robot)

# Uncomment these two lines to produce an excel file
#df = robot.get_dataframe(with_geo=False)
#df.to_excel("{basename}.xls")
#df.to_csv("{basename}.csv")
"""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Activated at the end of the with statement."""
        self.write()

    def add_text(self, text: str) -> None:
        """Add `text` to the script."""
        self.pytext += "\n" + text

    def write(self) -> None:
        """
        Write python script and json file with the list of files in the Robot.
        """
        with open(self.filepath_py, "wt") as fh:
            fh.write(self.pytext)
        make_executable(self.filepath_py)

        with open(self.filepath_json, "wt") as fh:
            fh.write(self.robot.to_json())



#!/usr/bin/env python
r"""
Luminescent properties (1D-CCM)
===============================

This example shows how to post-process the results of a LumiWork
following a one-dimensional configuration coordinate model (1D-CCM).
Based on a Eu-doped phosphor SrAlLi3N4:Eu presenting two non-equivalent sites for Eu
See example/flows/run_lumi_Eu_doped_SLA.py
"""

#%%
# Read the 4 points netcdf file produced by the two LumiWork

from abipy.lumi.deltaSCF import DeltaSCF
import abipy.data as abidata
import pandas as pd
import abipy.abilab as abilab

SLA_site_1 = DeltaSCF.from_four_points_file([abidata.ref_file("site_1_relaxed_gs_out_GSR.nc"),
                                            abidata.ref_file("site_1_unrelaxed_ex_out_GSR.nc"),
                                            abidata.ref_file("site_1_relaxed_ex_out_GSR.nc"),
                                            abidata.ref_file("site_1_unrelaxed_gs_out_GSR.nc")])

SLA_site_2 = DeltaSCF.from_four_points_file([abidata.ref_file("site_2_relaxed_gs_out_GSR.nc"),
                                            abidata.ref_file("site_2_unrelaxed_ex_out_GSR.nc"),
                                            abidata.ref_file("site_2_relaxed_ex_out_GSR.nc"),
                                            abidata.ref_file("site_2_unrelaxed_gs_out_GSR.nc")])

#%%
# To plot the luminescence lineshape following the one effective phonon mode model at 0K

SLA_site_1.plot_lineshape_1D_zero_temp(energy_range=[0.5,3],max_m=20,phonon_width=0.02,
               with_omega_cube='True',
               normalized='Area');

#%%
# To get a panda dataframe with the main restults :
dataframes=[]
df_1=SLA_site_1.get_dataframe('Site_1')
df_2=SLA_site_2.get_dataframe('Site_2')

pd.concat([df_1,df_2])


#%%
# To plot the magnitude of the displacements
SLA_site_1.plot_delta_R_distance(defect_symbol="Eu")

# %%
# To draw the configuration coordinates diagram.
SLA_site_1.draw_displaced_parabolas(scale_eff_freq=2.5);
SLA_site_2.draw_displaced_parabolas(scale_eff_freq=2.5);


# %%



"""
This module contains objects for analyzing
the PATH.nc file with the e-ph matrix elements along a k/q path
"""
from __future__ import annotations

import dataclasses
import numpy as np
import pandas as pd
import abipy.core.abinit_units as abu

from monty.string import marquee #, list_strings
from monty.functools import lazy_property
from monty.termcolor import cprint
from abipy.core.structure import Structure
from abipy.core.kpoints import Kpath
from abipy.core.mixins import AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.tools.typing import PathLike
#from abipy.tools.numtools import BzRegularGridInterpolator, nparr_to_df
from abipy.tools.plotting import (add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_axlims, set_visible,
    rotate_ticklabels, ax_append_title, set_ax_xylabels, linestyles, Marker, set_grid_legend)
#from abipy.tools import duck
from abipy.electrons.ebands import ElectronBands, RobotWithEbands
from abipy.dfpt.phonons import PhononBands
from abipy.dfpt.phtk import NonAnalyticalPh
from abipy.tools.typing import Figure
from abipy.abio.robots import Robot
from abipy.eph.common import BaseEphReader


class GpathFile(AbinitNcFile, Has_Structure, NotebookWriter):
    """
    This file stores the e-ph matrix elements along a k/q path
    and provides methods to analyze and plot results.

    Usage example:

    .. code-block:: python

        with GpathFile("out_GPATH.nc") as gpath:
            print(gpath)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GpathFile
    """

    @classmethod
    def from_file(cls, filepath: PathLike) -> GpathFile:
        """Initialize the object from a netcdf file."""
        return cls(filepath)

    def __init__(self, filepath: PathLike):
        super().__init__(filepath)
        self.r = GpathReader(filepath)

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    @lazy_property
    def ebands(self) -> ElectronBands:
        """Electron bands on the k+q path"""
        return self.r.read_ebands()

    @lazy_property
    def phbands(self) -> PhononBands:
        """Phonon bands on nq_path points."""
        return self.r.read_phbands()

    def close(self) -> None:
        """Close the file."""
        self.r.close()

    @lazy_property
    def params(self) -> dict:
        """dict with the convergence parameters, e.g. ``nbsum``."""
        #od = OrderedDict([
        #    ("nbsum", self.nbsum),
        #    ("nqibz", self.r.nqibz),
        #])
        ## Add EPH parameters.
        #od.update(self.r.common_eph_params)

        od = {}
        return od

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose: int=0) -> str:
        """String representation with verbosiy level ``verbose``."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))

        app("")
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))

        #app(f"nsppol: {self.r.nsppol}")
        #app(f"gstore_cplex: {self.r.cplex}")
        #app(f"gstore_kptopt: {self.r.kptopt}")
        #app(f"gstore_qptopt: {self.r.qptopt}")

        return "\n".join(lines)

    @add_fig_kwargs
    def plot_g_qpath(self, band_range=None, with_q=False, average_mode="all", ax_mat=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot ...

        Args:
            band_range:
            with_q
            average_mode:
            ax_mat: List of |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles
        """
        nrows, ncols = 2, self.r.nsppol
        #nrows, ncols = 3, self.r.nsppol
        ax_mat, fig, plt = get_axarray_fig_plt(ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=False, sharey=False, squeeze=False)

        qnorms = np.array([qpt.norm for qpt in self.phbands.qpoints]) if with_q else \
                 np.ones(len(self.phbands.qpoints))
        #band_range = (self.r.bstart, self.r.bstop) if band_range is None else band_range

        for spin in range(self.r.nsppol):
            g2_nuq, g2_nuq_unsym = self.r.get_g2nuq_average_spin(spin, band_range, average_mode)

            ax = ax_mat[0, spin]
            for mode in range(self.r.natom3):
                ax.plot(g2_nuq[mode] * qnorms, label=f"sym {mode=}")

            self.phbands.decorate_ax(ax, units="meV")
            set_grid_legend(ax, fontsize, xlabel=r"Wavevector $\mathbf{q}$", ylabel=r"$|g|$")

            ax = ax_mat[1, spin]
            for mode in range(self.r.natom3):
                ax.plot(g2_nuq_unsym[mode] * qnorms, label=f"unsym {mode=}")

            self.phbands.decorate_ax(ax, units="meV")
            set_grid_legend(ax, fontsize, xlabel=r"Wavevector $\mathbf{q}$", ylabel=r"$|g|$")

            #ax.plot(self.r.phfreqs_ha * abu.Ha_meV) # , label=f"e_kq")
            #set_grid_legend(ax, fontsize, xlabel=r"$\mathbf{q}$ wavevector", ylabel=r"$\omega_{\mathbf{q}\nu}$ (meV)")

            #ax = ax_mat[2, spin]
            #self.phbands.plot(ax=ax, show=False)
            #self.ebands.plot(ax=ax, show=False)
            #ax.plot(self.r.all_eigens_kq[spin]) # .transpose()) #, label=f"
            #set_grid_legend(ax, fontsize, xlabel=r"$\mathbf{k+q}$  wavevector")

        return fig

    #@add_fig_kwargs
    #def plot_g_kpath(self, band_range=None, with_q=False, average_mode="all", ax_mat=None, fontsize=8, **kwargs) -> Figure:

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.ebands.plot(show=False)
        if self.r.eph_fix_korq == "k":
            yield self.phbands.plot(show=False)
            yield self.plot_g_qpath()
        #if self.r.eph_fix_korq == "q":
        #    yield self.plot_g_kpath()

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("gpath = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(gpath)"),
            nbv.new_code_cell("gpath.ebands.plot();"),
            nbv.new_code_cell("gpath.phbands.plot();"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class GpathReader(BaseEphReader):
    """
    Reads data from file and constructs objects.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GpathReader
    """
    def __init__(self, filepath: PathLike):
        super().__init__(filepath)

        # Read important dimensions.
        self.nsppol = self.read_dimvalue("nsppol")
        self.natom = self.read_dimvalue("natom")
        self.natom3 = self.read_dimvalue("natom3")
        self.nb_in_g = self.read_dimvalue("nb_in_g")
        self.nk_path = self.read_dimvalue("nk_path")
        self.nq_path = self.read_dimvalue("nq_path")

        # eigens are in Ha, phfreq are in eV for historical reason
        self.phfreqs_ha = self.read_value("phfreqs")
        self.all_eigens_k = self.read_value("all_eigens_k")
        self.all_eigens_kq = self.read_value("all_eigens_kq")

        # Read important variables.
        self.eph_fix_korq = self.read_string("eph_fix_korq")
        self.eph_fix_wavec = self.read_value("eph_fix_wavevec")
        #self.completed = self.read_value("gstore_completed")

        # Note conversion Fortran --> C for the bstart index.
        nband = self.read_dimvalue("nband")
        self.bstart = self.read_value("bstart") - 1
        self.bstop = self.read_value("bstop")
        self.band_range = [self.bstart, self.bstop]

    def read_ebands(self):
        """
        Overrides method of superclass as we cannot rely of the etsf-io file format,
        and we have to build the ebands manually.
        """
        structure = self.read_structure()

        nspinor = self.read_dimvalue("nspinor")
        nspden = self.read_dimvalue("nspden")
        nelect = self.read_value("nelect")
        fermie = self.read_value("fermie") * abu.Ha_eV

        kpath_frac_coords = self.read_value("kpoints")
        qpath_frac_coords = self.read_value("qpoints")
        frac_coords = kpath_frac_coords + qpath_frac_coords
        path = Kpath(structure.lattice.reciprocal_lattice, frac_coords, weights=None, names=None, ksampling=None)

        # eigens are in Ha
        if self.nq_path > 1:
            all_eigens = self.read_value("all_eigens_kq") * abu.Ha_eV
        else:
            all_eigens = self.read_value("all_eigens_k") * abu.Ha_eV

        occfacts = np.zeros_like(all_eigens)

        return ElectronBands(structure, path, all_eigens, fermie, occfacts, nelect, nspinor, nspden)

    def read_phbands(self) -> PhononBands:
        """
        Read the phonon band structure along the q-path.
        """
        amu_list = self.read_value("atomic_mass_units")
        atomic_numbers = self.read_value("atomic_numbers")
        amu = {at: a for at, a in zip(atomic_numbers, amu_list)}

        # phfreqs are in eV for historical reason
        phfreqs = self.read_value("phfreqs")
        # Complex array with the Cartesian displacements in **Angstrom**
        phdispl_cart = self.read_value("phdispl_cart", cmode="c") * abu.Bohr_Ang

        structure = self.read_structure()
        qpath_frac_coords = self.read_value("qpoints")

        path_qq = Kpath(structure.lattice.reciprocal_lattice, qpath_frac_coords, weights=None, names=None, ksampling=None)

        non_anal_ph = NonAnalyticalPh.from_ncreader(self) if "non_analytical_directions" in self.rootgrp.variables else None

        return PhononBands(structure=structure, qpoints=path_qq, phfreqs=phfreqs, phdispl_cart=phdispl_cart, amu=amu,
                           non_anal_ph=non_anal_ph,
                           # TODO ?
                           #epsinf=epsinf,
                           #zcart=zcart,
                           )

    def get_g2nuq_average_spin(self, spin: int, band_range: list|tuple, average_mode: str, eps_mev: float=0.01):
        """
        Average ...

        Args:
            spin: Spin index
            band_range:
            average_mode:
            eps_mev: Tolerance in meV used to detect degeneracies.
        """
        # Consistency check
        if self.nk_path != 1:
            raise ValueError(f"{self.nk_path=} != 1. In this case, one cannot ask for q-dependent g(k,q)!")

        eps_ha = eps_mev / abu.Ha_meV
        eps_ev = eps_ha * abu.Ha_eV

        nsppol, natom3 = self.nsppol, self.natom3
        # Number of m, n bands in g_mn, the first band starts at bstart.
        nb_in_g = self.nb_in_g
        bstart = self.bstart

        all_eigens_k = self.all_eigens_k    # eV units
        all_eigens_kq = self.all_eigens_kq  # eV units
        phfreqs_ha = self.phfreqs_ha        # Ha units

        # Now read the e-ph matrix elements. On disk we have
        #                                                  n-index, m-index
        # double gkq2_nu(nsppol, nk_path, nq_path, natom3, nb_in_g, nb_in_g) ;
		#  gkq2_nu:_FillValue = -1. ;
        #
        # with nk_path == 1
        #                                      m-index, n-index
        # In memory we want: (nq_path, natom3, nb_in_g, nb_in_g)

        absg = np.sqrt(self.read_variable("gkq2_nu")[spin, 0][:].transpose(0, 1, 3, 2).copy()) * abu.Ha_meV
        absg_unsym = absg.copy()

        # Average over phonons.
        absg_sym = np.zeros_like(absg)

        for iq in range(self.nq_path):
            for nu in range(natom3):
                # Find all mu where |w_1 - w_nu| < eps_ha
                mask = np.abs(phfreqs_ha[iq, :] - phfreqs_ha[iq, nu]) < eps_ha
                # Sum the squared values of absg over the selected mu indices
                g2_mn = np.sum(absg[iq, mask, :, :]**2, axis=0)
                # Compute the symmetrized value (nn is the number of valid matches)
                nn = np.sum(mask)
                absg_sym[iq, nu, :, :] = np.sqrt(g2_mn / nn)

        # Average over k electrons.
        # MG FIXME: Note the difference with a similar function in gkq here I use absg and not absgk
        absg = absg_sym.copy()
        g2_nu = np.zeros((natom3), dtype=float)
        for iq in range(self.nq_path):
            for jbnd in range(nb_in_g):
                for ibnd in range(nb_in_g):
                    w_1 = all_eigens_k[spin, 0, ibnd]
                    g2_nu[:], nn = 0.0, 0
                    for pbnd in range(nb_in_g):
                        w_2 = all_eigens_k[spin, 0, pbnd]
                        if abs(w_2 - w_1) >= eps_ev: continue
                        nn += 1
                        g2_nu += absg[iq,:,jbnd,pbnd] ** 2
                    absg_sym[iq,:,jbnd,ibnd] = np.sqrt(g2_nu / nn)

        # Average over k electrons.
        # MG FIXME: Note the difference with a similar function in gkq here I use absg and not absgk
        #absg = absg_sym.copy()
        #for iq in range(self.nq_path):
        #    for jbnd in range(nb_in_g):
        #        for ibnd in range(nb_in_g):
        #            w_1 = all_eigens_k[spin, 0, ibnd]
        #            # Create mask to find all pbnd where |w_2 - w_1| < eps_ev
        #            mask = np.abs(all_eigens_k[spin, 0, :] - w_1) < eps_ev
        #            # Sum the squared values of absg over the selected pbnd indices (nn is the number of matching energy levels)
        #            nn = np.sum(mask)
        #            g2_nu = np.sum(absg[iq, :, jbnd, mask] ** 2, axis=-1)
        #            print(f"{mask=}")
        #            print(f"{absg[iq, :, jbnd, mask].shape=}")
        #            print(f"{absg.shape=}")
        #            print(f"{mask.shape=}")
        #            print(f"{g2_nu.shape=}")
        #            # Compute the symmetrized absg
        #            absg_sym[iq, :, jbnd, ibnd] = np.sqrt(g2_nu / nn)

        # Average over k+q electrons.
        absg = absg_sym.copy()
        for iq in range(self.nq_path):
          for ibnd in range(nb_in_g):
              for jbnd in range(nb_in_g):
                  w_1 = all_eigens_kq[spin, iq, jbnd]
                  g2_nu[:], nn = 0.0, 0
                  for pbnd in range(nb_in_g):
                      w_2 = all_eigens_kq[spin, iq, pbnd]
                      if abs(w_2 - w_1) >= eps_ev: continue
                      nn += 1
                      g2_nu += absg[iq,:,pbnd,ibnd] ** 2
                  absg_sym[iq,:,jbnd,ibnd] = np.sqrt(g2_nu / nn)

        # (nq_path, natom3, nb_in_g, nb_in_g) -> (natom3, nq_path, nb_in_g, nb_in_g)
        absg_sym = absg_sym.transpose(1, 0, 2, 3).copy()
        absg_unsym = absg_unsym.transpose(1, 0, 2, 3).copy()
        print(f"{absg_unsym.shape=}")

        # Take the trace over the last two axes
        #absg_sym = np.trace(absg_sym, axis1=-2, axis2=-1)
        #absg_unsym = np.trace(absg_unsym, axis1=-2, axis2=-1)

        # sum_{mn}
        absg_sym = np.sum(absg_sym, axis=(-2, -1)) / nb_in_g**2
        absg_unsym = np.sum(absg_unsym, axis=(-2, -1)) / nb_in_g**2

        return absg_sym, absg_unsym


class GpathRobot(Robot, RobotWithEbands):
    """
    This robot analyzes the results contained in multiple GPATH.nc files.

    Usage example:

    .. code-block:: python

        robot = GpathRobot.from_files([
            "t04o_GPATH.nc",
            "t05o_GPATH.nc",
            ])


    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GstoreRobot
    """
    EXT = "GPATH"

    #def neq(self, ref_basename: str | None = None, verbose: int = 0) -> int:
    #    """
    #    Compare all GPATHE.nc files stored in the robot
    #    """
    #    # Find reference gstore. By default the first file in the robot is used.
    #    ref_gstore = self._get_ref_abifile_from_basename(ref_basename)

    #    exc_list = []
    #    ierr = 0
    #    for other_gstore in self.abifiles:
    #        if ref_gstore.filepath == other_gstore.filepath:
    #            continue
    #        print("Comparing: ", ref_gstore.basename, " with: ", other_gstore.basename)
    #        try:
    #            ierr += self._neq_two_gstores(ref_gstore, other_gstore, verbose)
    #            cprint("EQUAL", color="green")
    #        except Exception as exc:
    #            exc_list.append(str(exc))

    #    for exc in exc_list:
    #        cprint(exc, color="red")

    #    return ierr

    #@staticmethod
    #def _neq_two_gstores(gstore1: GstoreFile, gstore2: GstoreFile, verbose: int) -> int:
    #    """
    #    Helper function to compare two GSTORE files.
    #    """
    #    # These quantities must be the same to have a meaningfull comparison.
    #    aname_list = ["structure", "nsppol", "cplex", "nkbz", "nkibz",
    #                  "nqbz", "nqibz", "completed", "kzone", "qzone", "kfilter", "gmode",
    #                  "brange_spin", "erange_spin", "glob_spin_nq", "glob_nk_spin",
    #                 ]

    #    for aname in aname_list:
    #        self._compare_attr_name(aname, gstore1, gstore2)

    #    # Now compare the gkq objects for each spin.
    #    ierr = 0
    #    for spin in range(gstore1.nsppol):
    #        gqk1, gqk2 = gstore1.gqk_spin[spin], gstore2.gqk_spin[spin]
    #        ierr += gqk1.neq(gqk2, verbose)

    #    return ierr

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        #for fig in self.get_ebands_plotter().yield_figs(): yield fig

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporary file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.GstoreRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            #nbv.new_code_cell("ebands_plotter = robot.get_ebands_plotter()"),
        ])

        # Mixins
        #nb.cells.extend(self.get_baserobot_code_cells())
        #nb.cells.extend(self.get_ebands_code_cells())

        return self._write_nb_nbpath(nb, nbpath)



from __future__ import annotations

import pytest
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def itest_tolsymerror_handler(fwp):
    """
    Test the handler of TolSymError. The test triggers:

        --- !TolSymError
        message: |
            Could not find the point group
        src_file: symptgroup.F90
        src_line: 236
        ...

    at the level of the symmetry finder and autoparal fails
    because it cannot find the parallel configurations.
    """
    pytest.xfail("tolsymerror_handler has been disabled because this problem has been fixed in v9.")
    structure = dict(
        acell=(1.0, 1.0, 1.0),
        xred=[
           1.0001907690, 1.0040151117, 0.0099335191,
           0.2501907744, 0.2540150788, 0.2599335332],
        rprim=[
          -6.2733366562, 0.0000000000, -3.6219126071,
          -6.2733366562, 0.0000000000,  3.6219126071,
          -4.1822244376, 5.9145585205,  0.0000000000],
        typat=(1, 1),
        ntypat=1,
        znucl=14,
        natom=2,
    )

    inp = abilab.AbinitInput(structure=structure, pseudos=abidata.pseudos("14si.pspnc"))

    inp.set_vars(
         ntime=5,
         tolrff=0.02,
         shiftk=[0, 0, 0],
         ngkpt=(4, 4, 4),
         chksymbreak=0,
         ecut=4,
         tolmxf=5e-05,
         nshiftk=1,
         #tolsym=1e-10,
    )

    flow = flowtk.Flow(workdir=fwp.workdir, manager=fwp.manager)
    flow.register_task(inp, task_class=flowtk.RelaxTask)

    flow.allocate()
    assert flow.make_scheduler().start() == 0

    flow.show_status()
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    task = flow[0][0]
    assert len(task.corrections) == 1
    assert task.corrections[0]["event"]["@class"] == "TolSymError"


def itest_dilatmxerror_handler(fwp):
    """Test the handler of DilatmxError. The test triggers:

        --- !DilatmxError
        message: |
            Dilatmx has been exceeded too many times (4)
            Restart your calculation from larger lattice vectors and/or a larger dilatmx
        src_file: mover.F90
        src_line: 840
        ...

    in variable cell structural optimizations.
    """
    #if fwp.on_travis:
    pytest.xfail("dilatmxerror_handler is not portable and it's been disabled!")

    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    structure.scale_lattice(structure.volume * 0.8)
    # Perturb the structure (random perturbation of 0.1 Angstrom)
    #structure.perturb(distance=0.1)

    inp = abilab.AbinitInput(structure=structure, pseudos=abidata.pseudos("14si.pspnc"))

    inp.set_vars(
        ecut=4,
        ngkpt=[4, 4, 4],
        shiftk=[0, 0, 0],
        nshiftk=1,
        chksymbreak=0,
        paral_kgb=1,
        optcell=1,
        ionmov=2,
        ecutsm=0.5,
        dilatmx=1.01,
        tolrff=0.02,
        tolmxf=5.0e-5,
        strfact=100,
        ntime=50,
        #ntime=5, To test the restart
        )

    # Create the flow
    flow = flowtk.Flow(fwp.workdir, manager=fwp.manager)
    flow.register_task(inp, task_class=flowtk.RelaxTask)

    flow.allocate()
    assert flow.make_scheduler().start() == 0

    flow.show_status()
    if not flow.all_ok:
        flow.debug()
        raise RuntimeError()

    task = flow[0][0]
    # Don't check the number of corrections as it's not portable.
    assert len(task.corrections)
    for i in range(task.num_corrections):
        assert task.corrections[i]["event"]["@class"] == "DilatmxError"
    #assert 0



#!/usr/bin/env python
r"""
Effective masses with finite difference
=======================================

Flow to compute effective masses with finite difference methods.
Derivatives are computed along lines in k-space.
"""

import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def make_scf_input(nspinor=1, usepaw=0):
    """Returns GS input file for Silicon"""
    if nspinor == 1:
        pseudos = abidata.pseudos("14si.pspnc") if usepaw == 0 else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")
    else:
        pseudos = abidata.pseudos("Si_r.psp8") if usepaw == 0 else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")

    structure = dict(
         ntypat=1,
         natom=2,
         typat=[1, 1],
         znucl=14,
         #acell=3 * [10.26310667319252], # https://docs.abinit.org/tests/v7/Input/t82.in
         acell=3 * [10.2073557], # 5.4015 Ang
         rprim=[[0.0,  0.5,  0.5],
                [0.5,  0.0,  0.5],
                [0.5,  0.5,  0.0]],
         xred=[ [0.0 , 0.0 , 0.0],
                [0.25, 0.25, 0.25]],
    )

    # Build input
    scf_input = abilab.AbinitInput(structure=structure, pseudos=pseudos)

    # Set variables
    scf_input.set_vars(
        ecut=12,
        nband=8 if nspinor == 1 else 16,
        nspinor=nspinor,
        tolvrs=1e-8,
    )

    if scf_input.ispaw:
        scf_input.set_vars(pawecutdg=2 * scf_input["ecut"])

    scf_input.set_kmesh(ngkpt=[8, 8, 8], shiftk=[0, 0, 0])

    return scf_input


def build_flow(options):
    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Get the SCF input (here NC with SOC)
    scf_input = make_scf_input(nspinor=2, usepaw=0)

    # Build the flow with different steps.
    from abipy.flowtk.effmass_works import EffMassLineWork
    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    # Multiple calculations with different step for finite difference.
    # Use ndivsm != 0 at the first iteration to compute band structure along
    for i, step in enumerate((0.01, 0.005)):
        if i == 0: den_node = None
        work = EffMassLineWork.from_scf_input(scf_input, k0_list=(0, 0, 0),
                                              step=step, npts=10,
                                              red_dirs=[[1, 0, 0], [1, 1, 0]],
                                              #cart_dirs=[[1, 0, 0], [1, 1, 1], [1, 1, 0]],
                                              ndivsm=-20 if i == 0 else 0,
                                              den_node=den_node)

        # Will start from the DEN file produced in the first iteration.
        if i == 0: den_node = work[0]
        flow.register_work(work)

    return flow


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_effmass_finitediff -s



"""
Global variables used to initialize AbiPy environment in notebooks.
"""
from __future__ import annotations

from monty.termcolor import cprint

import os
import tempfile


__IN_NOTEBOOK = False


def in_notebook() -> bool:
    """True if we are running inside a jupyter notebook (and enable_notebook has been called)."""
    return __IN_NOTEBOOK


def disable_notebook() -> None:
    """Set ``in_notebook`` flag to False."""
    global __IN_NOTEBOOK
    __IN_NOTEBOOK = False


def enable_notebook(with_seaborn=True) -> None:
    """
    Set ``in_notebook`` flag to True and activate seaborn settings for notebooks if ``with_seaborn``.
    """
    global __IN_NOTEBOOK
    __IN_NOTEBOOK = True

    # Use seaborn settings for plots (optional)
    if with_seaborn:
        import seaborn as sns
        sns.set(context='notebook', style='darkgrid', palette='deep',
                font='sans-serif', font_scale=1, color_codes=False, rc=None)


def get_abinb_workdir() -> str:
    """
    Return the absolute path of the scratch directory used to produce
    and save temporary files when we are runnning inside a jupyter_ notebook.

    .. note:

        Due to web-browser policy, files used in the notebook must be within the current working directory.
    """
    wdir = os.path.join(os.getcwd(), "__abinb_workdir__")
    if not os.path.exists(wdir): os.mkdir(wdir)
    return wdir


def abinb_mkstemp(force_abinb_workdir=False, use_relpath=False, **kwargs) -> tuple:
    """
    Invoke mkstep with kwargs, return the (fd, name) of the temporary file.
    kwargs are passed to ``mkstemp`` except for ``dir`` if we are inside a jupyter notebook.

    Args:
        use_abipy_nbworkdir:
        use_relpath: Return relative path (os.path.relpath) if True else absolute (default)
            Relative paths are required if we are gonna use the temporary file in
            notebooks or in web browers.
            In this case, the caller is responsbile for calling the function with the correct flag.

    .. example:

        _, filename = abinb_mkstep(suffix="." + ext, text=True)
    """
    if in_notebook() or force_abinb_workdir:
        d = kwargs.pop("dir", None)
        if d is not None:
            cprint("Files should be created inside abipy_nbworkdir if we are inside jupyter or force_abinb_workdir",
                   "yellow")
        fd, path = tempfile.mkstemp(dir=get_abinb_workdir(), **kwargs)
    else:
        fd, path = tempfile.mkstemp(**kwargs)

    if use_relpath:
        path = os.path.relpath(path)

    return fd, path


def get_workdir(workdir: str) -> str:
    """
    Return temporary directory if workdir is None else workdir.
    Allow users to specify the temporary directory via ABIPY_TMPDIR env variable.
    """
    if workdir is None:
        dir = os.getenv("ABIPY_TMPDIR", default=None)
        import tempfile
        workdir = tempfile.mkdtemp(dir=dir)

    return workdir



#!/usr/bin/env python
r"""
Raman with independent-particle approximation (optic)
=====================================================

This script shows how to perform a Raman calculation with excitonic effects
included with the BSE formalism.
"""

import sys
import os
import numpy as np

import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk


def build_flow(options):
    # Get the unperturbed structure.
    pseudos = abidata.pseudos("14si.pspnc")
    base_structure = abidata.structure_from_ucell("Si")

    ngkpt = [6, 6, 6]

    etas = [-.001, 0, +.001]
    ph_displ = np.reshape(np.zeros(3*len(base_structure)), (-1,3))
    ph_displ[0,:] = [+1, 0, 0]
    ph_displ[1,:] = [-1, 0, 0]

    # Build new structures by displacing atoms according to the phonon displacement
    # ph_displ (in cartesian coordinates). The Displacement is normalized so that
    # the maximum atomic diplacement is 1 Angstrom and then multiplied by eta.
    modifier = abilab.StructureModifier(base_structure)

    displaced_structures = modifier.displace(ph_displ, etas, frac_coords=False)

    # Initialize flow. Each workflow in the flow defines a complete BSE calculation for given eta.
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(options.workdir, manager=options.manager)

    # Generate the different shifts to average
    ndiv = 1
    shift1D = np.arange(1, 2*ndiv+1, 2) / (2 * ndiv)
    all_shifts = [[x, y, z] for x in shift1D for y in shift1D for z in shift1D]

    for structure, eta in zip(displaced_structures, etas):
        for ishift,shift in enumerate(all_shifts):
            flow.register_work(raman_work(structure, pseudos, ngkpt, shift),
                               workdir="eta_" + str(eta) + "shift_" + str(ishift))

    return flow


def raman_work(structure, pseudos, ngkpt, shiftk):
    # Generate 3 different input files for computing optical properties with Optic.

    global_vars = dict(
        istwfk="*1",
        paral_kgb=0,
        ecut=8,
        nstep=200,
        diemac=12,
        ixc=7,
        chksymbreak=0,
        #accesswff=3
    )

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=5)
    multi.set_vars(global_vars)
    multi.set_kmesh(ngkpt=ngkpt, shiftk=shiftk)

    # GS run
    multi[0].set_vars(
        tolvrs=1e-8,
        nband=20,
        nbdbuf=2,
    )

    # Note kptopt 2 in NSCF and DDK
    # In principle kptopt 2 is needed only in DDK.
    # one could do a first NSCF run with kptopt 1, reread with kptopt 2 and enter DDK...

    # NSCF run
    multi[1].set_vars(
       iscf=-2,
       nband=40,
       nbdbuf=5,
       kptopt=2,
       tolwfr=1.e-12,
    )

    # DDK along 3 directions
    # Third dataset : ddk response function along axis 1
    # Fourth dataset : ddk response function along axis 2
    # Fifth dataset : ddk response function along axis 3
    for idir in range(3):
        rfdir = 3 * [0]
        rfdir[idir] = 1

        multi[2 + idir].set_vars(
           iscf=-3,
           nband=40,
           nbdbuf=5,
           nstep=1,
           nline=0,
           prtwf=3,
           kptopt=2,
           nqpt=1,
           qpt=[0.0, 0.0, 0.0],
           rfdir=rfdir,
           rfelfd=2,
           tolwfr=1.e-12,
        )

    scf_inp, nscf_inp, ddk1, ddk2, ddk3 = multi.split_datasets()
    ddk_inputs = [ddk1, ddk2, ddk3]

    work = flowtk.Work()
    scf_t = work.register_scf_task(scf_inp)
    nscf_t = work.register_nscf_task(nscf_inp, deps={scf_t: "DEN"})

    ddk_nodes = []
    for inp in ddk_inputs:
        ddk_t = work.register_ddk_task(inp, deps={nscf_t: "WFK"})
        ddk_nodes.append(ddk_t)

    optic_input = abilab.OpticInput(
        broadening=0.002,     # Value of the smearing factor, in Hartree
        domega=0.0003,        # Frequency mesh.
        maxomega=0.3,
        scissor=0.000,        # Scissor shift if needed, in Hartree
        tolerance=0.002,      # Tolerance on closeness of singularities (in Hartree)
        num_lin_comp=6,       # Number of components of linear optic tensor to be computed
        lin_comp=(11, 12, 13, 22, 23, 33), # Linear coefficients to be computed (x=1, y=2, z=3)
        num_nonlin_comp=0     # Number of components of nonlinear optic tensor to be computed
        #nonlin_comp=(123, 222),
    )

    optic_t = flowtk.OpticTask(optic_input, nscf_node=nscf_t, ddk_nodes=ddk_nodes)
    work.register(optic_t)

    return work


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""Tools to compute speed of sound."""
from __future__ import annotations

import math
import numpy as np
import pandas as pd
import abipy.core.abinit_units as abu

from pymatgen.core.units import bohr_to_angstrom, eV_to_Ha
from abipy.core.structure import Structure
from abipy.core.mixins import Has_Structure, NotebookWriter
from abipy.dfpt.ddb import DdbFile
from abipy.dfpt.phonons import PhononBands, get_dyn_mat_eigenvec, match_eigenvectors
from abipy.abio.inputs import AnaddbInput
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, set_visible, get_fig_plotly, get_figs_plotly, \
    add_plotly_fig_kwargs, PlotlyRowColDesc
from abipy.tools.typing import Figure



class SoundVelocity(Has_Structure, NotebookWriter):
    """
    Compute the speed of sound by fitting phonon frequencies along selected directions
    by linear least-squares fit.
    """
    def __init__(self, directions, sound_velocities, mode_types, structure,
                 labels=None, phfreqs=None, qpts=None):
        """
        Args:
            directions: list of qpoints identifying the directions for the calculation
                of the speed of sound. In fractional coordinates.
            sound_velocities: array of shape (len(directions), 3) with the values of the
                sound velocities in SI units.
            mode_types: array of shape (len(directions), 3) with the type of modes (transverse,
                longitudinal). None if not known.
            structure: |Structure| object.
            labels: list with the same length as direction with labels for the directions.
            phfreqs: array with shape (len(directions), 3, num_points) with the acoustic phonon
                frequencies along the directions.
            qpts: array with shape (len(directions), num_points, 3) with the coordinates of
                the qpoints in fractional used to fit the phonon frequencies.
        """
        self.directions = np.array(directions)
        self.sound_velocities = np.array(sound_velocities)
        self.mode_types = mode_types
        self.labels = labels
        self.phfreqs = np.array(phfreqs) if phfreqs else None
        self.qpts = np.array(qpts) if qpts else None
        self._structure = structure

    @property
    def structure(self) -> Structure:
        """|Structure| object"""
        return self._structure

    @property
    def n_directions(self) -> int:
        """Number of directions."""
        return len(self.directions)

    @classmethod
    def from_ddb(cls, ddb_path: str, 
                 directions=None, labels=None, num_points=20, qpt_norm=0.1,
                 ignore_neg_freqs=True, asr=2, chneut=1, dipdip=1, dipquad=1, quadquad=1,
                 ngqpt=None, spell_check=True, anaddb_kwargs=None, verbose=0, mpi_procs=1, workdir=None, manager=None,
                 return_input=False) -> SoundVelocity:
        """
        Creates and instance of the object. Runs anaddb along the specified
        directions or the standard directions in the standard paths given
        in :cite:`Setyawan2010`. The values of the speed of sound
        will be calculated as the slope of the linear fits along each direction.

        Args:
            ddb_path (str): path to the ddb file.
            directions (list): list of qpoints identifying the directions for the calculation
                of the speed of sound. In fractional coordinates.
            labels (list): list of string with the name of the directions.
            num_points (int): number of points calculated along each direction and used to
                fit the speed of sound.
            qpt_norm (float): Norm of the largest q-point in fractional coordinates for
                each of the directions considered.
            ignore_neg_freqs (bool): if True, q-points with negative frequencies will not be
                considered in the fit, in order to ignore inaccuracies in the long range
                behavior.
            asr, chneut, dipdip: Anaddb input variable. See official documentation.
            dipquad, quadquad: 1 to include DQ, QQ terms (provided DDB contains dynamical quadrupoles).
            ngqpt: Number of divisions for the q-mesh in the DDB file. Auto-detected if None (default).
            anaddb_kwargs: additional kwargs for anaddb.
            verbose: verbosity level. Set it to a value > 0 to get more information.
            mpi_procs: Number of MPI processes to use.
            workdir: Working directory. If None, a temporary directory is created.
            manager: |TaskManager| object. If None, the object is initialized from the configuration file.
            return_input: True if the |AnaddbInput| object should be returned as 2nd argument

        Returns: an instance of SoundVelocity
        """
        with DdbFile(ddb_path) as ddb:
            if ngqpt is None: ngqpt = ddb.guessed_ngqpt

            inp = AnaddbInput(ddb.structure, comment="ANADDB input for speed of sound",
                              anaddb_kwargs=anaddb_kwargs, spell_check=spell_check)

            q1shft = [[0, 0, 0]]
            inp.set_vars(
                ifcflag=1,
                ngqpt=np.array(ngqpt),
                q1shft=q1shft,
                nqshft=len(q1shft),
                asr=asr,
                chneut=chneut,
                dipdip=dipdip,
                dipquad=dipquad,
                quadquad=quadquad,
            )

            if not directions:
                hs = ddb.structure.hsym_kpath
                kpath = hs.kpath

                directions = []
                labels = []

                for chunk in kpath["path"]:
                    for i, q in enumerate(chunk):
                        if "Gamma" in q:
                            if i > 0 and q not in labels:
                                new_q = kpath["kpoints"][chunk[i - 1]]
                                directions.append(new_q)
                                labels.append(chunk[i - 1])
                            if i < len(chunk) - 1 and q not in labels:
                                new_q = kpath["kpoints"][chunk[i + 1]]
                                directions.append(new_q)
                                labels.append(chunk[i + 1])

            qpts = []
            for q in directions:
                q = qpt_norm * q / np.linalg.norm(q)
                steps = q / num_points
                qpts.extend((steps[:, None] * np.arange(num_points)).T)

            n_qpoints = len(qpts)
            qph1l = np.zeros((n_qpoints, 4))

            qph1l[:, :-1] = qpts
            qph1l[:, -1] = 1

            inp['qph1l'] = qph1l.tolist()
            inp['nph1l'] = n_qpoints

            task = ddb._run_anaddb_task(inp, mpi_procs=mpi_procs, workdir=workdir, manager=manager,
                                        verbose=verbose)

            phbst_path = task.outpath_from_ext("PHBST")

            new = cls.from_phbst(phbst_path, ignore_neg_freqs=ignore_neg_freqs, labels=labels)
            return new if not return_input else (new, inp)

    @classmethod
    def from_phbst(cls, phbst_path: str,
                   ignore_neg_freqs: bool = True,
                   labels: list[str] = None) -> SoundVelocity:
        """
        Creates an instance of the object starting interpolating the acoustic frequencies
        from a PHBST netcdf file.
        The file should contain a series of directions starting from gamma and with the
        same number of points for each direction, as the one produced in the from_ddb method.

        Args:
            phbst_path: path to the PHBST netcdf file.
            ignore_neg_freqs (bool): if True points with negative frequencies will not be
                considered in the fit, in order to ignore inaccuracies in the long range
                behavior.
            labels (list): list of string with the name of the directions.

        Returns: an instance of SoundVelocity
        """
        phb = PhononBands.from_file(phbst_path)
        structure = phb.structure

        rlatt = structure.lattice.reciprocal_lattice
        # q points in cartesian coordinate in 1/bohr, the original units are 1/A
        qpt_cart_coords = [rlatt.get_cartesian_coords(c) * bohr_to_angstrom for c in
                           phb.qpoints.frac_coords]
        qpt_cart_norms = np.linalg.norm(qpt_cart_coords, axis=1)

        # find the indices of the gamma points
        gamma_ind = []
        for i, q in enumerate(phb.qpoints.frac_coords):
            if np.array_equal(q, [0,0,0]):
                gamma_ind.append(i)

        n_directions = len(gamma_ind)

        n_points = len(phb.qpoints) / n_directions
        if not n_points.is_integer():
            raise ValueError('Error extracting information from {}'.format(phbst_path))
        n_points = int(n_points)

        phfreqs = phb.phfreqs
        eigdisp = phb.phdispl_cart

        sound_velocities = []
        mode_types = []
        directions = []

        all_acoustic_freqs = []
        all_qpts = []

        for i in range(n_directions):
            start = n_points * i
            # index of the end point used for the slice
            # (the position of the last point is actually end-1)
            end = n_points * (i + 1)
            dir_freqs = phfreqs[start: end]
            dir_displ = eigdisp[start: end]

            # matching bands
            dir_eigv = get_dyn_mat_eigenvec(dir_displ, structure, amu=phb.amu)
            n_freqs = 3 * len(structure)
            ind_match = np.zeros((n_points, n_freqs), dtype=int)
            ind_match[0] = range(n_freqs)

            for j in range(1, n_points):
                k = j - 1
                match = match_eigenvectors(dir_eigv[k], dir_eigv[j])
                ind_match[j] = [match[m] for m in ind_match[k]]

            acoustic_freqs = (dir_freqs[np.arange(n_points)[:, None], ind_match])[:, 0:3]
            acoustic_displ = (dir_displ[np.arange(n_points)[:, None], ind_match])[:, 0:3]

            direction = phb.qpoints[end - 1].frac_coords
            direction = np.array(direction) / np.linalg.norm(direction)
            directions.append(direction)

            # identify the first (not gamma) qpoint with all positive frequencies
            first_positive_freq_ind = None
            for j in range(1, n_points):
                if min(acoustic_freqs[j]) > 0:
                    first_positive_freq_ind = j
                    break

            if first_positive_freq_ind is None or first_positive_freq_ind - n_points / 2 > 0:
                raise ValueError("Too many negative frequencies along direction {}".format(direction))

            sv = []
            mt = []

            cart_versor = qpt_cart_coords[end - 1] / np.linalg.norm(qpt_cart_coords[end - 1])
            for k in range(3):
                start_fit = 0
                if ignore_neg_freqs and first_positive_freq_ind > 1:
                    start_fit = first_positive_freq_ind
                slope, se, _, _ = np.linalg.lstsq(qpt_cart_norms[start+start_fit:end][:, np.newaxis],
                                                  acoustic_freqs[start_fit:, k] * eV_to_Ha, rcond=None)
                sv.append(slope[0] * abu.velocity_at_to_si)

                # identify the type of the mode (longitudinal/transversal) based on the
                # scalar product between the eigendisplacement and the direction.
                disp_0 = acoustic_displ[first_positive_freq_ind + 1, k, 0:3]
                disp_0 = disp_0 / np.linalg.norm(disp_0)

                scalar_prod = np.abs(np.dot(disp_0, cart_versor))
                if scalar_prod > 0.9:
                    mt.append("longitudinal")
                elif scalar_prod < 0.1:
                    mt.append("transversal")
                else:
                    mt.append(None)

            # sort the lists based on the sound velocities
            sv, mt, freqs = zip(*sorted(zip(sv, mt, acoustic_freqs.T.tolist())))

            sound_velocities.append(sv)
            mode_types.append(mt)
            all_acoustic_freqs.append(freqs)
            all_qpts.append(phb.qpoints.frac_coords[start:end])

        return cls(directions=directions, sound_velocities=sound_velocities, mode_types=mode_types,
                   structure=structure, labels=labels, phfreqs=all_acoustic_freqs, qpts=all_qpts)

    def get_dataframe(self) -> pd.DataFrame:
        """
        Return a |pandas-DataFrame| with the data of the speed of sound.
        """
        columns = ["direction", "label", "sound velocity (m/s)", "mode type"]
        rows = []
        for i in range(self.n_directions):
            for m in range(3):
                rows.append([
                    tuple(np.round(self.directions[i], decimals=3)),
                    self.labels[i] if self.labels else "",
                    self.sound_velocities[i][m],
                    self.mode_types[i][m]
                ])

        return pd.DataFrame(rows, columns=columns).set_index(["direction", "label"])

    @add_fig_kwargs
    def plot_fit_freqs_dir(self, idir, ax=None, units="eV", fontsize=8, **kwargs) -> Figure:
        """
        Plots the phonon frequencies with matplotlib, if available, along the specified direction.
        The line representing the fitted value will be shown as well.

        Args:
            idir: index of the direction.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz"). Case-insensitive.
            fontsize: fontsize for subtitles

        Returns: |matplotlib-Figure|
        """
        if self.phfreqs is None or self.qpts is None:
            raise ValueError("The plot requires phonon frequencies and qpoints.")

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        ax.margins(x=0, y=0)
        rlatt = self.structure.lattice.reciprocal_lattice
        freqs = self.phfreqs[idir]
        qpt_cart_coords = np.array([np.linalg.norm(rlatt.get_cartesian_coords(c)) for c in self.qpts[idir]])
        slope = self.sound_velocities[idir] / abu.velocity_at_to_si * bohr_to_angstrom / eV_to_Ha

        units_factor = abu.phfactor_ev2units(units)

        title = "[{:.3f}, {:.3f}, {:.3f}]".format(*self.directions[idir])
        if self.labels:
            title += " - {}".format(self.labels[idir])

        for i, c in enumerate(["r", "b", "g"]):
            ax.scatter(qpt_cart_coords, freqs[i] * units_factor, color=c, marker="x")
            ax.plot(qpt_cart_coords, slope[i] * qpt_cart_coords * units_factor, color=c, ls="-")

        ax.set_title(title, fontsize=fontsize)
        ax.grid(True)
        ax.set_xlabel("Wave Vector")
        ax.set_ylabel(abu.wlabel_from_units(units))

        return fig

    @add_plotly_fig_kwargs
    def plotly_fit_freqs_dir(self, idir, fig=None, rcd=None, units="eV", fontsize=12, **kwargs):
        """
        Plots the phonon frequencies with plotly, if available, along the specified direction.
        The line representing the fitted value will be shown as well.

        Args:
            idir: index of the direction.
            fig: |plotly.graph_objects.Figure|
            rcd: PlotlyRowColDesc object used to specify the (row, col) of the subplot in the grid.
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz"). Case-insensitive.
            fontsize: fontsize for subtitles

        Returns: |plotly.graph_objects.Figure|
        """
        if self.phfreqs is None or self.qpts is None:
            raise ValueError("The plot requires phonon frequencies and qpoints.")

        title = "[{:.3f}, {:.3f}, {:.3f}]".format(*self.directions[idir])
        if self.labels:
            title += " - {}".format(self.labels[idir])

        rcd = PlotlyRowColDesc.from_object(rcd)
        ply_row, ply_col = rcd.ply_row, rcd.ply_col
        xaxis = 'xaxis%u' % rcd.iax
        yaxis = 'yaxis%u' % rcd.iax

        if fig is None:
            fig, _ = get_fig_plotly()
            fig.layout = dict(annotations=[dict(text=title, font_size=fontsize, x=0.5, xref='paper', xanchor='center',
                              y=1, yref='paper', yanchor='bottom' ,showarrow=False)],
                              yaxis_title_text=abu.wlabel_from_units(units, unicode=True),
                              xaxis_title_text="Wave Vector")
        else:
            fig.layout.annotations[idir].text = title
            fig.layout.annotations[idir].font.size = fontsize
            if idir == self.n_directions - 1:
                fig.layout[xaxis].title.text = "Wave Vector"
            if idir == 0:
                fig.layout[yaxis].title.text = abu.wlabel_from_units(units, unicode=True)

        fig.layout[xaxis].rangemode = 'tozero'
        fig.layout[yaxis].rangemode = 'tozero'

        rlatt = self.structure.lattice.reciprocal_lattice
        freqs = self.phfreqs[idir]
        qpt_cart_coords = np.array([np.linalg.norm(rlatt.get_cartesian_coords(c)) for c in self.qpts[idir]])
        slope = self.sound_velocities[idir] / abu.velocity_at_to_si * bohr_to_angstrom / eV_to_Ha

        units_factor = abu.phfactor_ev2units(units)

        for i, c in enumerate(["red", "blue", "green"]):
            fig.add_scatter(x=qpt_cart_coords, y=slope[i] * qpt_cart_coords * units_factor, line_color=c,
                          name='', showlegend=False, mode='lines', row=ply_row, col=ply_col)
            fig.add_scatter(x=qpt_cart_coords, y=freqs[i] * units_factor, marker=dict(symbol=4, size=8, color=c),
                          name='', showlegend=False, mode='markers', row=ply_row, col=ply_col)

        return fig

    @add_fig_kwargs
    def plot(self, units="eV", fontsize=8, **kwargs) -> Figure:
        """
        Plots the phonon frequencies with matplotlib, if available, along all the directions.
        The lines representing the fitted values will be shown as well.

        Args:
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz"). Case-insensitive.
            fontsize: fontsize for subtitles

        Returns: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=None)
        from matplotlib.gridspec import GridSpec

        nrows, ncols = math.ceil(self.n_directions / 2),  2
        gspec = GridSpec(nrows=nrows, ncols=ncols, wspace=0.15, hspace=0.25)

        for i in range(self.n_directions):
            axi = plt.subplot(gspec[i])
            self.plot_fit_freqs_dir(i, ax=axi, units=units, fontsize=fontsize, show=False)
            if i != self.n_directions - 1:
                set_visible(axi, False, "xlabel")
            if i != 0:
                set_visible(axi, False, "ylabel")

        return fig

    @add_plotly_fig_kwargs
    def plotly(self, units="eV", fontsize=12, **kwargs):
        """
        Plots the phonon frequencies with plotly, if available, along all the directions.
        The lines representing the fitted values will be shown as well.

        Args:
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz"). Case-insensitive.
            fontsize: fontsize for subtitles

        Returns: |plotly.graph_objects.Figure|
        """

        nrows, ncols = math.ceil(self.n_directions / 2),  2
        fig, _ = get_figs_plotly(nrows=nrows, ncols=ncols, subplot_titles=list(range(1, self.n_directions+1)),
                                 horizontal_spacing=0.05)

        for i in range(self.n_directions):
            rcd = PlotlyRowColDesc(i // 2, i % 2, nrows, ncols)
            self.plotly_fit_freqs_dir(i, fig, rcd, units=units, fontsize=fontsize, show=False)

        return fig

    def yield_figs(self, **kwargs):   # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        for i in range(self.n_directions):
            yield self.plot_fit_freqs_dir(i)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write an jupyter_ notebook to nbpath. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        tmpfile = self.pickle_dump()

        nb.cells.extend([
            nbv.new_code_cell("sv = abilab.SoundVelocity.pickle_load('{}')".format(tmpfile)),
            nbv.new_code_cell("sv.get_dataframe()")
        ])
        if self.phfreqs is not None and self.qpts is not None:
            for i in range(self.n_directions):
                nb.cells.append(nbv.new_code_cell("sv.plot_fit_freqs_dir({});".format(i)))

        return self._write_nb_nbpath(nb, nbpath)



#!/usr/bin/env python
"""Analyze the parallel efficiency of the SIGMA code (one shot G0W0 with plasmon-pole model and gwpara==2)"""
import sys
import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_inputs(paw=False):
    pseudos = abidata.pseudos("14si.pspnc", "8o.pspnc") if not paw else \
              abidata.pseudos("Si.GGA_PBE-JTH-paw.xml", "o.paw")

    structure = abidata.structure_from_ucell("SiO2-alpha")

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=4)

    ecut = 24
    multi.set_vars(
        ecut=ecut,
        pawecutdg=ecut*2 if paw else None,
        paral_kgb=0,
        istwfk="*1",
        timopt=-1,
    )

    multi.set_kmesh(ngkpt=[4,4,3], shiftk=[0.0, 0.0, 0.0])

    gs, nscf, scr, sigma = multi.split_datasets()

    # Dataset 1 (GS run)
    gs.set_vars(tolvrs=1e-6,
                nband=28,)

    # Dataset 2 (NSCF run)
    nscf.set_vars(iscf=-2,
                  tolwfr=1e-4,
                  nband=600,
                  nbdbuf=200,
                  )

    # Dataset3: Calculation of the screening.
    scr.set_vars(
        optdriver=3,
        gwpara=2,
        nband=25,
        ecutwfn=ecut,
        symchi=1,
        inclvkb=0,
        ecuteps=6.0,
    )

    # Dataset4: Calculation of the Self-Energy matrix elements (GW corrections)
    sigma.set_vars(
        optdriver=4,
        gwpara=2,
        ecutwfn=ecut,
        ecuteps=6.0,
        ecutsigx=ecut,
        symsigma=1,
        gw_qprange=1,
    )

    return gs, nscf, scr, sigma


def build_flow(options):
    """
    Build an `AbinitWorkflow` used for benchmarking ABINIT.
    """
    gs_inp, nscf_inp, scr_inp, sigma_inp = make_inputs(paw=options.paw)
    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    bands = flowtk.BandStructureWork(gs_inp, nscf_inp)
    flow.register_work(bands)
    flow.exclude_from_benchmark(bands)

    scr_work = flowtk.Work()
    scr_work.register_scr_task(scr_inp, deps={bands.nscf_task: "WFK"})
    flow.register_work(scr_work)
    flow.exclude_from_benchmark(scr_work)

    mpi_list = options.mpi_list

    for nband in [200, 400, 600]:
        sigma_work = flowtk.Work()

        if options.mpi_list is None:
            # Cannot call autoparal here because we need a WFK file.
            print("Using hard coded values for mpi_list")
            mpi_list = [np for np in range(1, nband+1) if abs(nband % np) < 1]
        if options.verbose: print("Using nband %d and mpi_list: %s" % (nband, mpi_list))

        for mpi_procs, omp_threads in product(mpi_list, options.omp_list):
            if not options.accept_mpi_omp(mpi_procs, omp_threads): continue
            inp = sigma_inp.new_with_vars(nband=nband)
            manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
            sigma_work.register_sigma_task(inp, manager=manager,
                                           deps={bands.nscf_task: "WFK", scr_work[0]: "SCR"})
        flow.register_work(sigma_work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Dielectric function with LFE
============================

This examples shows how to plot the macroscopic dielectric function
computed in the GW code (optdriver 3)
"""
import abipy.data as abidata
from abipy.abilab import abiopen

with abiopen(abidata.ref_file("sio2_SCR.nc")) as ncfile:
    # The SCR file contains a structure and electron bands in the IBZ.
    # We can thus use the ebands object to plot bands + DOS.
    print(ncfile)

    edos = ncfile.ebands.get_edos()
    ncfile.ebands.plot_with_edos(edos, title="KS energies used to compute the SCR file.")

    # sphinx_gallery_thumbnail_number = 2
    ncfile.plot_emacro(title="Macroscopic dielectric function of $SiO_2$ with local-field effects.")

    ncfile.plot_eelf(title="Electron Energy Loss Function of $SiO_2$")



"""Panels to interact with the SIGEPH file."""
from __future__ import annotations

import param
import panel as pn
import panel.widgets as pnw

from abipy.eph.sigeph import SigEPhFile
from .core import PanelWithElectronBands, mpl, ply, depends_on_btn_click #, PanelWithEbandsRobot


class SigEPhFilePanel(PanelWithElectronBands):
    """
    Panel with widgets to interact with a |SigEphFile|.
    """
    def __init__(self, sigeph: SigEPhFile, **params):
        PanelWithElectronBands.__init__(self, ebands=sigeph.ebands, **params)
        self.sigeph = sigeph

        self.sigma_spin_select = pnw.Select(name="Spin index", options=list(range(sigeph.nsppol)))
        self.sigma_kpoint_select = pnw.Select(name="Kpoint in Sigma_nk",
                options={"[%d]: %s" % (ik, repr(k)): ik for ik, k in enumerate(sigeph.sigma_kpoints)})

        self.plot_qpsolution_btn = pnw.Button(name="Plot Sigma_nk", button_type='primary')
        #sigma_band_select  = param.ObjectSelector(default=0, objects=[0], doc="Band index in sigma_nk")

    def plot_lws(self) -> pn.GridSpec:
        # Insert results in grid.
        gspec = pn.GridSpec(sizing_mode='scale_width')
        for i, rta_type in enumerate(("serta", "mrta")):
            fig = self.sigeph.plot_lws_vs_e0(rta_type=rta_type, **self.mpl_kwargs)
            gspec[i, 0] = fig
            fig = self.sigeph.plot_tau_vtau(rta_type=rta_type, **self.mpl_kwargs)
            gspec[i, 1] = fig

        return gspec

    def plot_qpgaps(self) -> pn.GridSpec:
        # Insert results in grid.
        gspec = pn.GridSpec(sizing_mode='scale_width')

        for i, qp_kpt in enumerate(self.sigeph.sigma_kpoints):
            fig = self.sigeph.plot_qpgaps_t(qp_kpoints=qp_kpt, qp_type="qpz0", **self.mpl_kwargs)
            gspec[i, 0] = fig
            fig = self.sigeph.plot_qpgaps_t(qp_kpoints=qp_kpt, qp_type="otms", **self.mpl_kwargs)
            gspec[i, 1] = fig

        return gspec

    def plot_qps_vs_e0(self):
        return mpl(self.sigeph.plot_qps_vs_e0(**self.mpl_kwargs))

    @depends_on_btn_click('plot_qpsolution_btn')
    def on_plot_qpsolution_sk(self):
        fig = self.sigeph.plot_qpsolution_sk(self.sigma_spin_select.value, self.sigma_kpoint_select.value,
                                             itemp=0, with_int_aw=True,
                                             ax_list=None, xlims=None, fontsize=8, **self.mpl_kwargs)
        return mpl(fig)

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the SIGEPH file."""
        d = {}
        d["Summary"] = self.get_summary_view_for_abiobj(self.sigeph)
        #app(("e-Bands", pn.Row(self.get_plot_ebands_widgets(), self.on_plot_ebands_btn)))

        # Build different tabs depending on the calculation type.
        # TODO
        #if self.sigeph.imag_only:
        #    d["LWS"] = self.plot_lws

        #else:
        #    d["QP-gaps"] = self.plot_qpgaps
        #    d["QP_vs_e0"] = self.plot_qps_vs_e0
        #    d["QPSolution"] = pn.Row(
        #        pn.Column("## Quasiparticle solution",
        #                  self.sigma_spin_select,
        #                  self.sigma_kpoint_select,
        #                  self.plot_qpsolution_btn),
        #        self.on_plot_qpsolution_sk
        #    )

        #d["Structure"] = self.get_structure_view()
        d["NcFile"] = self.sigeph.get_ncfile_view()

        if as_dict: return d
        return self.get_template_from_tabs(d, template=kwargs.get("template", None))



"""
This module gathers the most important classes and helper functions used for scripting.
"""
import sys
import os
import collections

from itertools import chain
from typing import ClassVar, Optional, List, Union
from tabulate import tabulate

####################
### Monty import ###
####################
from monty.termcolor import cprint

#######################
### Pymatgen import ###
#######################
# Tools for unit conversion
import pymatgen.core.units as units
FloatWithUnit = units.FloatWithUnit
ArrayWithUnit = units.ArrayWithUnit

####################
### Abipy import ###
####################
from abipy.flowtk import Pseudo, PseudoTable, Mrgscr, Mrgddb, Flow, Work, TaskManager, AbinitBuild, flow_main
from abipy.core.release import __version__, min_abinit_version
from abipy.core.globals import enable_notebook, in_notebook, disable_notebook
from abipy.core.structure import (Lattice, Structure, StructureModifier, dataframes_from_structures,
  mp_match_structure, mp_search, cod_search, display_structure)
from abipy.core.mixins import TextFile, JsonFile, CubeFile
from abipy.core.func1d import Function1D
from abipy.core.kpoints import set_atol_kdiff
from abipy.abio.robots import Robot
from abipy.abio.inputs import AbinitInput, MultiDataset, AnaddbInput, OpticInput
from abipy.abio.abivars import AbinitInputFile
from abipy.abio.outputs import AbinitLogFile, AbinitOutputFile, OutNcFile, AboRobot
from abipy.tools.printing import print_dataframe
from abipy.tools.notebooks import print_source, print_doc
from abipy.tools.serialization import mjson_load, mjson_loads, mjson_write
from abipy.abio.factories import *
from abipy.electrons.ebands import (ElectronBands, ElectronBandsPlotter, ElectronDos, ElectronDosPlotter,
    dataframe_from_ebands, EdosFile)
from abipy.electrons.gsr import GsrFile, GsrRobot
from abipy.electrons.eskw import EskwFile
from abipy.electrons.psps import PspsFile, PspsRobot
from abipy.electrons.gw import SigresFile, SigresRobot
from abipy.electrons.gwr import GwrFile
from abipy.electrons.bse import MdfFile, MdfRobot
from abipy.electrons.scissors import ScissorsBuilder
from abipy.electrons.scr import ScrFile
from abipy.electrons.denpot import (DensityNcFile, VhartreeNcFile, VxcNcFile, VhxcNcFile, PotNcFile,
    DensityFortranFile, Cut3dDenPotNcFile)
from abipy.electrons.fatbands import FatBandsFile
from abipy.electrons.optic import OpticNcFile, OpticRobot
from abipy.electrons.fold2bloch import Fold2BlochNcfile
from abipy.dfpt.phonons import (PhbstFile, PhbstRobot, PhononBands, PhononBandsPlotter, PhdosFile, PhononDosPlotter,
    PhdosReader, phbands_gridplot)
from abipy.dfpt.ddb import DdbFile, DdbRobot
from abipy.dfpt.anaddbnc import AnaddbNcFile, AnaddbNcRobot
from abipy.dfpt.gruneisen import GrunsNcFile
from abipy.dynamics.hist import HistFile, HistRobot
from abipy.waves import WfkFile
from abipy.eph.a2f import A2fFile, A2fRobot
from abipy.eph.sigeph import SigEPhFile, SigEPhRobot
from abipy.eph.cumulant import CumulantEPhFile
from abipy.eph.varpeq import VarpeqFile
from abipy.eph.eph_plotter import EphPlotter
from abipy.eph.v1sym import V1symFile
from abipy.eph.gkq import GkqFile, GkqRobot
from abipy.eph.v1qnu import V1qnuFile
from abipy.eph.v1qavg import V1qAvgFile
from abipy.eph.rta import RtaFile, RtaRobot
from abipy.eph.transportfile import TransportFile
from abipy.eph.gstore import GstoreFile
from abipy.wannier90 import WoutFile, AbiwanFile, AbiwanRobot
from abipy.electrons.lobster import CoxpFile, ICoxpFile, LobsterDoscarFile, LobsterInput, LobsterAnalyzer

from abipy.dynamics.cpx import EvpFile
#try:
#    from abipy.ml.aseml import AseMdLog
#except ImportError:
#    AseMdLog = None

#from abipy.electrons.abitk import ZinvConvFile, TetraTestFile

# Abinit Documentation.
from abipy.abio.abivars_db import get_abinit_variables, abinit_help, docvar


def _straceback():
    """Returns a string with the traceback."""
    import traceback
    return traceback.format_exc()


# Abinit text files. Use OrderedDict for nice output in show_abiopen_exc2class.
ext2file = collections.OrderedDict([
    # ABINIT files
    (".abi", AbinitInputFile),
    (".in", AbinitInputFile),
    (".abo", AbinitOutputFile),
    (".out", AbinitOutputFile),
    (".log", AbinitLogFile),
    (".cif", Structure),
    (".abivars", Structure),
    (".ucell", Structure),
    (".cssr", Structure),
    (".json", JsonFile),
    (".py", TextFile),
    (".sh", TextFile),
    (".stdin", TextFile),
    (".stderr", TextFile),
    (".err", TextFile),
    (".files", TextFile),
    (".stdout", TextFile),
    (".cube", CubeFile),
    ("anaddb.nc", AnaddbNcFile),
    ("DEN", DensityFortranFile),
    (".wout", WoutFile),
    ("EDOS", EdosFile),
    # Pseudos
    (".psp8", Pseudo),
    (".pspnc", Pseudo),
    (".fhi", Pseudo),
    ("JTH.xml", Pseudo),
    (".upf", Pseudo),
    # Lobster files.
    ("COHPCAR.lobster", CoxpFile),
    ("COOPCAR.lobster", CoxpFile),
    ("ICOHPLIST.lobster", ICoxpFile),
    ("DOSCAR.lobster", LobsterDoscarFile),
    # Vasp files.
    ("POSCAR", Structure),
    (".vasp", Structure),
    # ASE files
    (".xyz", Structure),
    #("ZINVCONV.nc", ZinvConvFile),
    #("TETRATEST.nc", TetraTestFile),
    # QE/CP files
    (".evp", EvpFile),
    # ASE files produced by Abipy.
    #("md.aselog", AseMdLog),
])

# Abinit files require a special treatment.
abiext2ncfile = collections.OrderedDict([
    ("GSR.nc", GsrFile),
    ("ESKW.nc", EskwFile),
    ("DEN.nc", DensityNcFile),
    ("OUT.nc", OutNcFile),
    ("VHA.nc", VhartreeNcFile),
    ("VXC.nc", VxcNcFile),
    ("VHXC.nc", VhxcNcFile),
    ("POT.nc", PotNcFile),
    ("WFK.nc", WfkFile),
    ("HIST.nc", HistFile),
    ("PSPS.nc", PspsFile),
    ("DDB", DdbFile),
    ("PHBST.nc", PhbstFile),
    ("PHDOS.nc", PhdosFile),
    ("SCR.nc", ScrFile),
    ("SIGRES.nc", SigresFile),
    ("GWR.nc", GwrFile),
    ("GRUNS.nc", GrunsNcFile),
    ("MDF.nc", MdfFile),
    ("FATBANDS.nc", FatBandsFile),
    ("FOLD2BLOCH.nc", Fold2BlochNcfile),
    ("CUT3DDENPOT.nc", Cut3dDenPotNcFile),
    ("OPTIC.nc", OpticNcFile),
    ("A2F.nc", A2fFile),
    ("SIGEPH.nc", SigEPhFile),
    ("GSTORE.nc", GstoreFile),
    ("TRANSPORT.nc",TransportFile),
    ("RTA.nc",RtaFile),
    ("V1SYM.nc", V1symFile),
    ("GKQ.nc", GkqFile),
    ("V1QNU.nc", V1qnuFile),
    ("V1QAVG.nc", V1qAvgFile),
    ("ABIWAN.nc", AbiwanFile),
    ("EPH_CUMULANT.nc", CumulantEPhFile),
    ("VARPEQ.nc", VarpeqFile),
])


def abiopen_ext2class_table():
    """
    Print the association table between file extensions and File classes.
    """
    table = []
    for ext, cls in chain(ext2file.items(), abiext2ncfile.items()):
        table.append((ext, str(cls)))

    return tabulate(table, headers=["Extension", "Class"])


def extcls_supporting_panel(as_table=True, **tabulate_kwargs):
    """
    Return list of (ext, cls) tuples where ext is the file extension supporting the `get_panel` method
    i.e files that can be integrated with the panel dashboard and `cls` the python class associated to it.
    """
    items = []

    for ext, cls in chain(ext2file.items(), abiext2ncfile.items()):
        if hasattr(cls, "get_panel"): items.append((ext, str(cls)))

    if not as_table: return items

    return tabulate(items, headers=["Extension", "Class"], **tabulate_kwargs)


def abipanel(**kwargs):
    """
    Activate panel extensions used by AbiPy. Return panel module.
    """
    try:
        from abipy.panels.core import abipanel
        return abipanel(**kwargs)
    except ImportError as exc:
        cprint("use `conda install panel` or `pip install panel` to install the python package.", "red")
        raise exc


def abifile_subclass_from_filename(filename: str) -> ClassVar:
    """
    Returns the appropriate class associated to the given filename.
    """
    if os.path.basename(filename) == Flow.PICKLE_FNAME:
        return Flow

    from abipy.tools.text import rreplace
    for ext, cls in ext2file.items():
        # This to support gzipped files.
        if filename.endswith(".gz"): filename = rreplace(filename, ".gz", "", occurrence=1)
        if filename.endswith(ext): return cls

    ext = filename.split("_")[-1]
    try:
        return abiext2ncfile[ext]
    except KeyError:
        for ext, cls in abiext2ncfile.items():
            if filename.endswith(ext): return cls

    msg = f"""
abiopen cannot handle this file as no phython class has been registered for file:\n`{filename}`\n\n
This is the list of file extensions supported by abiopen:\n\n
{abiopen_ext2class_table()}
"""
    raise ValueError(msg)


def dir2abifiles(top: str, recurse: bool = True) -> dict:
    """
    Analyze the filesystem starting from directory `top` and
    return an ordered dictionary mapping the directory name to the list
    of files supported by ``abiopen`` contained within that directory.
    If not ``recurse``, children directories are not analyzed.
    """
    dl = collections.defaultdict(list)

    if recurse:
        for dirpath, dirnames, filenames in os.walk(top):
            for f in filenames:
                path = os.path.join(dirpath, f)
                if not isabifile(path): continue
                dl[dirpath].append(path)
    else:
        for f in os.listdir(top):
            path = os.path.join(top, f)
            if not isabifile(path): continue
            dl[top].append(path)

    return collections.OrderedDict([(k, dl[k]) for k in sorted(dl.keys())])


def isabifile(filepath: str) -> bool:
    """
    Return True if `filepath` can be opened with ``abiopen``.
    """
    try:
        abifile_subclass_from_filename(filepath)
        return True
    except ValueError:
        return False


def abiopen(filepath: str):
    """
    Factory function that opens any file supported by abipy.
    File type is detected from the extension

    Args:
        filepath: string with the filename.
    """
    # Handle ~ in filepath.
    filepath = os.path.expanduser(filepath)

    # Handle zipped files by creating temporary file with correct extension.
    root, ext = os.path.splitext(filepath)
    if ext in (".bz2", ".gz", ".z"):
        from monty.io import zopen
        with zopen(filepath, "rt") as f:
            import tempfile
            _, tmp_path = tempfile.mkstemp(suffix=os.path.basename(root), text=True)
            cprint("Creating temporary file: %s" % tmp_path, "yellow")
            with open(tmp_path, "wt") as t:
                t.write(f.read())
            filepath = tmp_path

    if os.path.basename(filepath) == "__AbinitFlow__.pickle":
        return Flow.pickle_load(filepath)

    # Handle old output files produced by Abinit.
    import re
    outnum = re.compile(r".+\.out[\d]+")
    abonum = re.compile(r".+\.abo[\d]+")
    if outnum.match(filepath) or abonum.match(filepath):
        return AbinitOutputFile.from_file(filepath)

    if os.path.basename(filepath) == "log":
        # Assume Abinit log file.
        return AbinitLogFile.from_file(filepath)

    if os.path.basename(filepath).endswith("phonopy_params.yaml"):
        # Handle phonopy object.
        import phonopy
        return phonopy.load(filepath)

    cls = abifile_subclass_from_filename(filepath)
    return cls.from_file(filepath)


def abirobot(filepaths: Union[str, List[str]]) -> Robot:
    """
    Factory function to create and return a Robot subclass from a list of filenames
    The Robot subclass is detected from the extension of the first file hence
    all files are assumed to have the same extension.

    Args:
        filepaths: List of strings with the filename.
    """
    from monty.string import list_strings
    filepaths = list_strings(filepaths)
    path = filepaths[0]
    idx = path.rfind("_")
    if idx == -1:
        raise ValueError("Cannot find `_` in the first string")
    ext = path[idx+1:]

    cls = Robot.class_for_ext(ext)
    robot = cls.from_files(filepaths)
    return robot


def software_stack(as_dataframe: bool = False):
    """
    Import all the hard dependencies and some optional packages.
    Returns ordered dict: package --> string with version info or pandas dataframe if as_dataframe.
    """
    import platform
    system, node, release, version, machine, processor = platform.uname()
    # These packages are required
    import numpy, scipy, netCDF4, pymatgen, apscheduler, pydispatch, plotly
    import ruamel.yaml as yaml

    from importlib import import_module

    def get_version(pkg_name):
        """Return version of package from string."""
        try:
            mod = import_module(pkg_name)
            return mod.__version__
        except Exception:
            return None

    try:
        from pymatgen.core import __version__ as pmg_version
        #from pymatgen.settings import __version__ as pmg_version
    except AttributeError:
        pmg_version = pymatgen.__version__

    d = collections.OrderedDict([
        ("system", system),
        ("python_version", platform.python_version()),
        ("numpy", numpy.version.version),
        ("scipy", scipy.version.version),
        ("netCDF4", netCDF4.__version__),
        ("apscheduler", apscheduler.version),
        ("pydispatch", pydispatch.__version__),
        ("ruamel.yaml", yaml.__version__),
        ("boken", get_version("bokeh")),
        ("panel", get_version("panel")),
        ("plotly", get_version("plotly")),
        ("ase", get_version("ase")),
        ("phonopy", get_version("phonopy")),
        ("monty", get_version("monty")),
        ("pymatgen", pmg_version),
        ("abipy", __version__),
    ])

    if not as_dataframe: return d
    import pandas as pd
    return pd.Series(data=d, name="version").to_frame().rename_axis("Package")


def abicheck(verbose: int = 0) -> str:
    """
    This function tests if the most important ABINIT executables
    can be found in $PATH and whether the python modules needed
    at run-time can be imported. Return string with error messages, empty if success.
    """

    err_lines = []
    app = err_lines.append

    try:
        manager = TaskManager.from_user_config()
    except Exception:
        manager = None
        app(_straceback())

    # Get info on the Abinit build.
    # This to avoid having to depend on pytest.
    #from abipy.core.testing import cmp_version
    def cmp_version(this: str, other: str, op: str = ">=") -> bool:
        """
        Compare two version strings with the given operator ``op``
        >>> assert cmp_version("1.1.1", "1.1.0") and not cmp_version("1.1.1", "1.1.0", op="==")
        """
        from packaging.version import parse as parse_version
        from monty.operator import operator_from_str
        op = operator_from_str(op)
        return op(parse_version(this), parse_version(other))

    from abipy.flowtk import PyFlowScheduler

    if manager is not None:
        cprint("AbiPy Manager:\n%s\n" % str(manager), color="green")
        build = AbinitBuild(manager=manager)
        if not build.has_netcdf: app("Abinit executable does not support netcdf")
        cprint("Abinitbuild:\n%s" % str(build), color="magenta")
        if verbose: cprint(str(build.info), color="magenta")
        print()
        if not cmp_version(build.version, min_abinit_version, op=">="):
            app("Abipy requires Abinit version >= %s but got %s" % (min_abinit_version, build.version))

    # Get info on the scheduler.
    try:
        scheduler = PyFlowScheduler.from_user_config()
        cprint("Abipy Scheduler:\n%s\n" % str(scheduler), color="yellow")
    except Exception as exc:
        app(_straceback())

    try:
        d = software_stack()
        cprint("Installed packages:", color="blue")
        cprint(tabulate(list(d.items()), headers=["Package", "Version"]), color="blue")
        print()
    except ImportError:
        app(_straceback())

    import pprint
    shell_vars = [os.environ.get(vname, "") for vname in ("PATH", "LD_LIBRARY_PATH", "TMPDIR")]
    print("Important Shell Variables:")
    pprint.pprint(shell_vars)

    return "\n".join(err_lines)


def install_config_files(workdir: Optional[str] = None, force_reinstall: Optional[bool] = False):
    """
    Install pre-defined configuration files for the TaskManager and the Scheduler
    in the workdir directory.

    Args:
        workdir: Directory when configuration files should be produced. Use ~/abinit/abipy/ if None
        force_reinstall: Allow overwrite pre-existent configuration files. By default, the function
            raises RuntimeError if configuration files are already present.
    """
    workdir = os.path.join(os.path.expanduser("~"), ".abinit", "abipy") if workdir is None else workdir
    print("Installing configuration files in directory:", workdir)
    from monty.os import makedirs_p
    makedirs_p(workdir)

    scheduler_path = os.path.join(workdir, "scheduler.yaml")
    scheduler_yaml = """
# The launcher will stop submitting jobs when the
# number of jobs in the queue is >= Max number of jobs
max_njobs_inqueue: 2

# Maximum number of cores that can be used by the scheduler.
max_ncores_used: 2

# number of hours to wait.
#hours: 0

# number of minutes to wait.
#minutes: 0

# number of seconds to wait.
seconds: 2

# Send mail to the specified address (accepts string or list of strings).
# PRO TIP: the scheduler WILL try to send and email after a default time of 4 days. If you
#          comment out the mailto address, this will cause the scheduler to terminate, with
#          potentially nefarious effects on your running jobs. If you do not wish to receive
#          emails, a work around is to set the variable `remindme_s` below to something very
#          large (say, 100 days).
#mailto: nobody@nowhere.com

# verbosity level (int, default 0)
#verbose: 0

# The scheduler will shutdown when the number of python exceptions is > max_num_pyexcs
#max_num_pyexcs: 2

# The scheduler will shutdown when the number of Abinit errors is > max_num_abierrs
#max_num_abierrs: 0

# The scheduler will shutdow when the total number of tasks launched is > safety_ratio * tot_num_tasks.
#safety_ratio: 5

# Send an e-mail to mailto every remindme_s seconds.
#remindme_s: 345600
"""

    manager_path = os.path.join(workdir, "manager.yaml")
    manager_yaml = """
qadapters:
    -
      priority: 1
      queue:
        qname: abipy
        qtype: shell
      job:
        mpi_runner: mpirun
        pre_run:
         - export OMP_NUM_THREADS=1
         # IMPORTANT: Change the below line so that the abinit executable is in PATH
         #- export PATH=$HOME/git_repos/abinit/_build/src/98_main:$PATH
         #- ulimit -s unlimited; ulimit -n 2048

      limits:
         min_cores: 1
         max_cores: 2
         timelimit: 0:10:0
      hardware:
         num_nodes: 1
         sockets_per_node: 1
         cores_per_socket: 2
         mem_per_node: 4 GB
"""

    # Write configuration files.
    if not os.path.isfile(scheduler_path) or force_reinstall:
        with open(scheduler_path, "wt") as fh:
            fh.write(scheduler_yaml)
        print("Scheduler configuration file written to:", scheduler_path)
    else:
        raise RuntimeError("Configuration file: %s already exists.\nUse force_reinstall option to overwrite it" % scheduler_path)

    if not os.path.isfile(manager_path) or force_reinstall:
        with open(manager_path, "wt") as fh:
            fh.write(manager_yaml)
        print("Manager configuration file written to:", manager_path)
    else:
        raise RuntimeError("Configuration file: %s already exists.\nUse force_reinstall option to overwrite it" % manager_path)

    print("""
Configuration files installed successfully.
Please edit the configuration options according to your installation.
In particular, edit the `pre_run` section in manager.yml
so that the abinit executable is in $PATH.
""")

    return 0


def abipy_logo1() -> str:
    """http://www.text-image.com/convert/pic2ascii.cgi"""
    return r"""
                 `:-                                                               -:`
         --`  .+/`                              `                                  `/+.  .-.
   `.  :+.   /s-                   `yy         .yo                                   -s/   :+. .`
 ./.  +o`   /s/           `-::-`   `yy.-::-`   `:-    .:::-`   -:`     .:`            /s/   :s- ./.
.o.  /o:   .oo.         .oyo++syo. `yyys++oys. -ys  -syo++sy+` sy-     +y:            .oo-   oo` `o.
++   oo.   /oo          yy-    -yy `yy:    .yy`-ys .ys`    /yo sy-     +y:             oo/   /o:  ++
+/   oo`   /oo         `yy.    .yy` yy.    `yy`-ys :ys     :yo oy/     oy:             +o/   :o:  /o
-/   :+.   -++`         -sy+::+yyy` .sy+::+yy- -ys :yys/::oys. `oyo::/syy:            `++-   /+.  /:
 --  `//    /+-           -/++/-//    -/++/-   `+: :yo:/++/.     .:++/:oy:            -+/   `+-  --
  `.`  -:    :/`                                   :yo                 +y:           `/:`  `:. `.`
        `..   .:.                                   .`                 `.           .:.  `..
                ...                                                               ...
"""


def abipy_logo2() -> str:
    """http://www.text-image.com/convert/pic2ascii.cgi"""
    return r"""
MMMMMMMMMMMMMMMMNhdMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMdhmMMMMMMMMMMMMMMM
MMMMMMMMMddNMMmoyNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMNyomMMmhmMMMMMMMM
MMMmmMMhomMMMy/hMMMMMMMMMMMMMMMMMMMN::MMMMMMMMMm:oMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMd+yMMMhomMmmMMM
MmsmMMs+NMMMy+yMMMMMMMMMMMNhyyhmMMMN::mhyyhmMMMNhdMMMMmhyydNMMMdyNMMMMMmyNMMMMMMMMMMMMy+yMMMh+dMmsmM
m+mMMy+hMMMd++mMMMMMMMMMm+:+ss+:+mMN:::/ss+:+mMd:/MMd/:+so/:oNM/:dMMMMMo:yMMMMMMMMMMMMm++dMMMo+NMN+m
osMMMo+mMMMy+oMMMMMMMMMM::dMMMMd:/MN::hMMMMd::Nd:/Mm:/NMMMMy:oM/:dMMMMMo:yMMMMMMMMMMMMMo+yMMMy+hMMso
oyMMMooNMMMyooMMMMMMMMMN::mMMMMm::NM::mMMMMN::Nd:/Mh:/MMMMMh:+Mo:yMMMMM+:yMMMMMMMMMMMMMooyMMMyohMMyo
dyMMMysmMMMdooNMMMMMMMMMd/:oyys:::NMd/:oyys::dMd:/Mh::/shyo:/mMN+:+yhs/::yMMMMMMMMMMMMNoodMMMysmMMyh
MddMMNyhMMMMysdMMMMMMMMMMMdyooydssMMMMdysoydMMMNsyMh:+hsosydMMMMMmysosho:yMMMMMMMMMMMMdsyMMMNydMMddM
MMNmNMMddMMMMhyNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMh:oMMMMMMMMMMMMMMMMMo:yMMMMMMMMMMMNyhMMMNhmMNmNMM
MMMMMMMMNmmMMMmhmMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMmNMMMMMMMMMMMMMMMMMNdMMMMMMMMMMMmhmMMNmmMMMMMMMM
MMMMMMMMMMMMMMMMmmNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMNmmMMMMMMMMMMMMMMM
"""


def abipy_logo3() -> str:
    """http://www.text-image.com/convert/pic2ascii.cgi"""
    return r"""\
             `-.                                                  `--`
      -:. `//`              `/.       ::                           `+: `-:`
 --``+:  `o+        `.--.`  -y/.--.`  ::   `---`  `-     ..         `o+  `o-`--
:/  /o   /o-       -oo/:+s/ -yy+:/os- ss .oo/:/s+`:y.    yo          :o:  :o. /:
o-  o/   oo`       ss    /y..y/    ss ss +y`   -y::y-    yo          -o/  .o- -o
:-  //   /+.       :s+--/sy- +s/--+s: ss oyo:-:so``os:-:oyo          -+:  -+. -/
 -` `/.  `/:        `-:::-:`  `-::-`  -- oy-:::.    .:::-yo          //`  :- `-
   `  ..` `:-                            :+              /:         --` `-` `
            `.`                                                   ..`
"""
